from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_google
from arcade_google.models import DocumentFormat, OrderBy
from arcade_google.tools import (
    get_file_tree_structure,
    search_and_retrieve_documents,
    search_documents,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_module(arcade_google)


@tool_eval()
def get_file_tree_structure_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="get my google drive's file tree structure including shared drives",
        user_message="get my google drive's file tree structure including shared drives",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": True,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="get my google drive's file tree structure without shared drives",
        user_message="get my google drive's file tree structure without shared drives",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": False,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="what are the files in the folder 'hello world' in my google drive?",
        user_message="what are the files in the folder 'hello world' in my google drive?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": False,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    suite.add_case(
        name="how many files are there in all my google drives, including shared ones?",
        user_message="how many files are there in all my google drives, including shared ones?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={
                    "restrict_to_shared_drive_id": None,
                    "include_shared_drives": True,
                    "include_organization_domain_documents": False,
                    "order_by": None,
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="restrict_to_shared_drive_id", weight=0.5 / 4),
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5 / 4),
            BinaryCritic(critic_field="order_by", weight=0.5 / 4),
            BinaryCritic(critic_field="limit", weight=0.5 / 4),
        ],
    )

    return suite


@tool_eval()
def search_documents_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Search documents in Google Drive",
        user_message="get my 49 most recently created documents, list the ones created most recently first.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "order_by": [OrderBy.CREATED_TIME_DESC.value],
                    "limit": 49,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="order_by", weight=0.5),
            BinaryCritic(critic_field="limit", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in Google Drive based on document keywords",
        user_message="Search the documents that contain the word 'greedy' and the phrase 'hello, world'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["greedy", "hello, world"],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=1.0),
        ],
    )

    suite.add_case(
        name="Search documents in a specific Google Drive based on document keywords",
        user_message="Search the documents that contain the word 'greedy' and the phrase 'hello, world' in the drive with id 'abc123'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["greedy", "hello, world"],
                    "search_only_in_shared_drive_id": "abc123",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="search_only_in_shared_drive_id", weight=0.5),
            BinaryCritic(critic_field="document_contains", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in a Google Drive Workspace organization domain based on document keywords",
        user_message="Search the documents that contain the phrase 'hello, world' in the organization domain",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "document_contains": ["hello, world"],
                    "include_organization_domain_documents": True,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_organization_domain_documents", weight=0.5),
            BinaryCritic(critic_field="document_contains", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search documents in shared drives",
        user_message="Search the 5 documents from all drives corpora that nobody has touched in forever, excluding shared drives.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_documents,
                args={
                    "limit": 5,
                    "include_shared_drives": False,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="include_shared_drives", weight=0.5),
            BinaryCritic(critic_field="limit", weight=0.5),
        ],
    )

    suite.add_case(
        name="No tool call case",
        user_message="List my 10 most recently modified documents that are stored in my Microsoft OneDrive.",
        expected_tool_calls=[],
        critics=[],
    )

    return suite


@tool_eval()
def search_and_retrieve_documents_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive search and retrieve tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive documents using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Search and retrieve (write summary)",
        user_message="Write a summary of the documents in my Google Drive about 'MX Engineering'",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["MX Engineering"],
                    "return_format": DocumentFormat.MARKDOWN,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (project proposal)",
        user_message="Display the document contents in HTML format from my Google Drive that contain the phrase 'project proposal'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["project proposal"],
                    "return_format": DocumentFormat.HTML,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (meeting notes)",
        user_message="Retrieve documents that contain both 'meeting notes' and 'budget' in JSON format.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["meeting notes", "budget"],
                    "return_format": DocumentFormat.GOOGLE_API_JSON,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=0.5),
            BinaryCritic(critic_field="return_format", weight=0.5),
        ],
    )

    suite.add_case(
        name="Search and retrieve (Q1 report)",
        user_message="Show me the content of the documents that mention 'Q1 report' but do not include the expression 'Project XYZ'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_and_retrieve_documents,
                args={
                    "document_contains": ["Q1 report"],
                    "document_not_contains": ["Project XYZ"],
                    "return_format": DocumentFormat.MARKDOWN,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="document_contains", weight=1 / 3),
            BinaryCritic(critic_field="document_not_contains", weight=1 / 3),
            BinaryCritic(critic_field="return_format", weight=1 / 3),
        ],
    )

    return suite



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import ToolExecutionError
from googleapiclient.errors import HttpError

from arcade_google_docs.tools import (
    create_blank_document,
    create_document_from_text,
    get_document_by_id,
    insert_text_at_end_of_document,
)
from arcade_google_docs.utils import build_docs_service


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.fixture
def mock_get_service():
    with patch("arcade_google_docs.tools.get." + build_docs_service.__name__) as mock_build_service:
        yield mock_build_service.return_value


@pytest.fixture
def mock_update_service():
    with patch(
        "arcade_google_docs.tools.update." + build_docs_service.__name__
    ) as mock_build_service:
        yield mock_build_service.return_value


@pytest.fixture
def mock_create_service():
    with patch(
        "arcade_google_docs.tools.create." + build_docs_service.__name__
    ) as mock_build_service:
        yield mock_build_service.return_value


@pytest.mark.asyncio
async def test_get_document_by_id_success(mock_context, mock_get_service):
    # Mock the service.documents().get().execute() method
    mock_get_service.documents.return_value.get.return_value.execute.return_value = {
        "body": {"content": [{"endIndex": 1, "paragraph": {}}]},
        "documentId": "test_document_id",
        "title": "Test Document",
    }

    result = await get_document_by_id(mock_context, "test_document_id")

    assert result["documentId"] == "test_document_id"
    assert result["title"] == "Test Document"


@pytest.mark.asyncio
async def test_get_document_by_id_http_error(mock_context, mock_get_service):
    # Simulate HttpError
    mock_get_service.documents.return_value.get.return_value.execute.side_effect = HttpError(
        resp=AsyncMock(status=404), content=b'{"error": {"message": "Not Found"}}'
    )

    with pytest.raises(ToolExecutionError, match="Error in execution of GetDocumentById"):
        await get_document_by_id(mock_context, "invalid_document_id")


@pytest.mark.asyncio
async def test_insert_text_at_end_of_document_success(mock_context, mock_update_service):
    # Mock get_document_by_id to return a document with endIndex
    with patch(
        "arcade_google_docs.tools.update.get_document_by_id",
        return_value={"body": {"content": [{"endIndex": 1, "paragraph": {}}]}},
    ):
        # Mock the service.documents().batchUpdate().execute() method
        mock_update_service.documents.return_value.batchUpdate.return_value.execute.return_value = {
            "documentId": "test_document_id",
            "replies": [],
        }

        result = await insert_text_at_end_of_document(
            mock_context, "test_document_id", "Sample text"
        )

        assert result["documentId"] == "test_document_id"


@pytest.mark.asyncio
async def test_insert_text_at_end_of_document_http_error(mock_context, mock_update_service):
    with patch(
        "arcade_google_docs.tools.update.get_document_by_id",
        return_value={"body": {"content": [{"endIndex": 1, "paragraph": {}}]}},
    ):
        # Simulate HttpError during batchUpdate
        mock_update_service.documents.return_value.batchUpdate.return_value.execute.side_effect = (
            HttpError(resp=AsyncMock(status=400), content=b'{"error": {"message": "Bad Request"}}')
        )

        with pytest.raises(
            ToolExecutionError, match="Error in execution of InsertTextAtEndOfDocument"
        ):
            await insert_text_at_end_of_document(mock_context, "test_document_id", "Sample text")


@pytest.mark.asyncio
async def test_create_blank_document_success(mock_context, mock_create_service):
    # Mock the service.documents().create().execute() method
    mock_create_service.documents.return_value.create.return_value.execute.return_value = {
        "documentId": "new_document_id",
        "title": "New Document",
    }

    result = await create_blank_document(mock_context, "New Document")

    assert result["documentId"] == "new_document_id"
    assert result["title"] == "New Document"
    assert "documentUrl" in result


@pytest.mark.asyncio
async def test_create_blank_document_http_error(mock_context, mock_create_service):
    # Simulate HttpError during create
    mock_create_service.documents.return_value.create.return_value.execute.side_effect = HttpError(
        resp=AsyncMock(status=403), content=b'{"error": {"message": "Forbidden"}}'
    )

    with pytest.raises(ToolExecutionError, match="Error in execution of CreateBlankDocument"):
        await create_blank_document(mock_context, "New Document")


@pytest.mark.asyncio
async def test_create_document_from_text_success(mock_context, mock_create_service):
    with patch(
        "arcade_google_docs.tools.create." + create_blank_document.__name__
    ) as mock_create_blank_document:
        # Mock create_blank_document
        mock_create_blank_document.return_value = {
            "documentId": "new_document_id",
            "title": "New Document",
        }

        # Mock the service.documents().batchUpdate().execute() method
        mock_create_service.documents.return_value.batchUpdate.return_value.execute.return_value = {
            "documentId": "new_document_id",
            "replies": [],
        }

        result = await create_document_from_text(mock_context, "New Document", "Hello, World!")

        assert result["documentId"] == "new_document_id"
        assert result["title"] == "New Document"
        assert "documentUrl" in result


@pytest.mark.asyncio
async def test_create_document_from_text_http_error(mock_context, mock_create_service):
    with patch(
        "arcade_google_docs.tools.create." + create_blank_document.__name__
    ) as mock_create_blank_document:
        # Mock create_blank_document
        mock_create_blank_document.return_value = {
            "documentId": "new_document_id",
            "title": "New Document",
        }

        # Simulate HttpError during batchUpdate
        mock_create_service.documents.return_value.batchUpdate.return_value.execute.side_effect = (
            HttpError(
                resp=AsyncMock(status=500), content=b'{"error": {"message": "Internal Error"}}'
            )
        )

        with pytest.raises(
            ToolExecutionError, match="Error in execution of CreateDocumentFromText"
        ):
            await create_document_from_text(mock_context, "New Document", "Hello, World!")



from typing import Annotated

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Spotify
from arcade_tdk.errors import RetryableToolError, ToolExecutionError

from arcade_spotify.tools.constants import RESPONSE_MSGS
from arcade_spotify.tools.models import Device, SearchType
from arcade_spotify.tools.search import search
from arcade_spotify.tools.utils import (
    convert_to_playback_state,
    get_url,
    send_spotify_request,
)


# NOTE: This tool only works for Spotify Premium users
@tool(requires_auth=Spotify(scopes=["user-read-playback-state", "user-modify-playback-state"]))
async def adjust_playback_position(
    context: ToolContext,
    absolute_position_ms: Annotated[
        int | None, "The absolute position in milliseconds to seek to"
    ] = None,
    relative_position_ms: Annotated[
        int | None,
        "The relative position from the current playback position in milliseconds to seek to",
    ] = None,
) -> Annotated[str, "Success/failure message"]:
    """Adjust the playback position within the currently playing track.

    Knowledge of the current playback state is NOT needed to use this tool as it handles
    clamping the position to valid start/end boundaries to prevent overshooting or negative values.

    This tool allows you to seek to a specific position within the currently playing track.
    You can either provide an absolute position in milliseconds or a relative position from
    the current playback position in milliseconds.

    Note:
        Either absolute_position_ms or relative_position_ms must be provided, but not both.
    """
    if (absolute_position_ms is None) == (relative_position_ms is None):
        raise RetryableToolError(
            "Either absolute_position_ms or relative_position_ms must be provided, but not both",
            additional_prompt_content=(
                "Provide a value for either absolute_position_ms or "
                "relative_position_ms, but not both."
            ),
            retry_after_ms=500,
        )

    if relative_position_ms is not None:
        playback_state = await get_playback_state(context)
        if playback_state.get("device_id") is None:
            return RESPONSE_MSGS["no_track_to_adjust_position"]

        absolute_position_ms = playback_state["progress_ms"] + relative_position_ms

    absolute_position_ms = max(0, absolute_position_ms or 0)

    url = get_url("player_seek_to_position")
    params = {"position_ms": absolute_position_ms}

    try:
        response = await send_spotify_request(context, "PUT", url, params=params)
    except httpx.HTTPStatusError as e:
        raise ToolExecutionError(f"Failed to adjust playback position: {e}") from e

    if response.status_code == 404:
        return RESPONSE_MSGS["no_track_to_adjust_position"]

    response.raise_for_status()

    return RESPONSE_MSGS["playback_position_adjusted"]


# NOTE: This tool only works for Spotify Premium users
@tool(requires_auth=Spotify(scopes=["user-read-playback-state", "user-modify-playback-state"]))
async def skip_to_previous_track(
    context: ToolContext,
) -> Annotated[str, "Success/failure message"]:
    """Skip to the previous track in the user's queue, if any"""
    url = get_url("player_skip_to_previous")

    response = await send_spotify_request(context, "POST", url)

    if response.status_code == 404:
        return RESPONSE_MSGS["no_track_to_go_back_to"]

    response.raise_for_status()

    return RESPONSE_MSGS["playback_skipped_to_previous_track"]


# NOTE: This tool only works for Spotify Premium users
@tool(requires_auth=Spotify(scopes=["user-read-playback-state", "user-modify-playback-state"]))
async def skip_to_next_track(
    context: ToolContext,
) -> Annotated[str, "Success/failure message"]:
    """Skip to the next track in the user's queue, if any"""
    url = get_url("player_skip_to_next")

    response = await send_spotify_request(context, "POST", url)

    if response.status_code == 404:
        return RESPONSE_MSGS["no_track_to_skip"]

    response.raise_for_status()

    return RESPONSE_MSGS["playback_skipped_to_next_track"]


# NOTE: This tool only works for Spotify Premium users
@tool(requires_auth=Spotify(scopes=["user-read-playback-state", "user-modify-playback-state"]))
async def pause_playback(
    context: ToolContext,
) -> Annotated[str, "Success/failure message"]:
    """Pause the currently playing track, if any"""
    playback_state = await get_playback_state(context)

    # There is no current state, therefore nothing to pause
    if playback_state.get("device_id") is None:
        return RESPONSE_MSGS["no_track_to_pause"]
    # Track is already paused
    if playback_state.get("is_playing") is False:
        return RESPONSE_MSGS["track_is_already_paused"]

    url = get_url("player_pause_playback")

    response = await send_spotify_request(context, "PUT", url)
    response.raise_for_status()

    return RESPONSE_MSGS["playback_paused"]


# NOTE: This tool only works for Spotify Premium users
@tool(
    requires_auth=Spotify(
        scopes=["user-read-playback-state", "user-modify-playback-state"],
    )
)
async def resume_playback(
    context: ToolContext,
) -> Annotated[str, "Success/failure message"]:
    """Resume the currently playing track, if any"""
    playback_state = await get_playback_state(context)

    # There is no current state, therefore nothing to resume
    if playback_state.get("device_id") is None:
        return RESPONSE_MSGS["no_track_to_resume"]
    # Track is already playing
    if playback_state.get("is_playing") is True:
        return RESPONSE_MSGS["track_is_already_playing"]

    url = get_url("player_modify_playback")

    response = await send_spotify_request(context, "PUT", url)
    response.raise_for_status()

    return RESPONSE_MSGS["playback_resumed"]


# NOTE: This tool only works for Spotify Premium users
@tool(
    requires_auth=Spotify(
        scopes=["user-read-playback-state", "user-modify-playback-state"],
    )
)
async def start_tracks_playback_by_id(
    context: ToolContext,
    track_ids: Annotated[
        list[str],
        "A list of Spotify track (song) IDs to play. Order of execution is not guarenteed.",
    ],
    position_ms: Annotated[
        int | None,
        "The position in milliseconds to start the first track from",
    ] = 0,
) -> Annotated[str, "Success/failure message"]:
    """Start playback of a list of tracks (songs)"""

    devices = [
        Device(**device) for device in (await get_available_devices(context)).get("devices", [])
    ]

    # If no active device is available, pick the first one.
    # Otherwise, Spotify defaults to the active device.
    device_id = None
    if devices and not any(device.is_active for device in devices):
        device_id = devices[0].id

    params = {"device_id": device_id} if device_id else {}

    url = get_url("player_modify_playback")
    body = {
        "uris": [f"spotify:track:{track_id}" for track_id in track_ids],
        "position_ms": position_ms,
    }

    response = await send_spotify_request(context, "PUT", url, params=params, json_data=body)

    if response.status_code == 404:
        return RESPONSE_MSGS["no_active_device"]

    response.raise_for_status()

    return RESPONSE_MSGS["playback_started"]


@tool(requires_auth=Spotify(scopes=["user-read-playback-state"]))
async def get_playback_state(
    context: ToolContext,
) -> Annotated[dict, "Information about the user's current playback state"]:
    """
    Get information about the user's current playback state,
    including track or episode, and active device.
    This tool does not perform any actions. Use other tools to control playback.
    """
    url = get_url("player_get_playback_state")
    response = await send_spotify_request(context, "GET", url)
    response.raise_for_status()
    data = {"is_playing": False} if response.status_code == 204 else response.json()
    return convert_to_playback_state(data).to_dict()


@tool(requires_auth=Spotify(scopes=["user-read-currently-playing"]))
async def get_currently_playing(
    context: ToolContext,
) -> Annotated[dict, "Information about the user's currently playing track"]:
    """Get information about the user's currently playing track"""
    url = get_url("player_get_currently_playing")
    response = await send_spotify_request(context, "GET", url)
    response.raise_for_status()
    data = {"is_playing": False} if response.status_code == 204 else response.json()
    return convert_to_playback_state(data).to_dict()


# NOTE: This tool only works for Spotify Premium users
@tool(
    requires_auth=Spotify(
        scopes=["user-read-playback-state", "user-modify-playback-state"],
    )
)
async def play_artist_by_name(
    context: ToolContext,
    name: Annotated[str, "The name of the artist to play"],
) -> Annotated[str, "Success/failure message"]:
    """Plays a song by an artist and queues four more songs by the same artist"""
    q = f"artist:{name}"
    search_results = await search(context, q, [SearchType.TRACK], 5)
    if not search_results["tracks"]["items"]:
        message = RESPONSE_MSGS["artist_not_found"].format(artist_name=name)
        raise RetryableToolError(
            message,
            additional_prompt_content=f"{message} Try a different artist name.",
            retry_after_ms=500,
        )
    track_ids = [item["id"] for item in search_results["tracks"]["items"]]
    response = await start_tracks_playback_by_id(context, track_ids)

    return str(response)


# NOTE: This tool only works for Spotify Premium users
@tool(
    requires_auth=Spotify(
        scopes=["user-read-playback-state", "user-modify-playback-state"],
    )
)
async def play_track_by_name(
    context: ToolContext,
    track_name: Annotated[str, "The name of the track to play"],
    artist_name: Annotated[str | None, "The name of the artist of the track"] = None,
) -> Annotated[str, "Success/failure message"]:
    """Plays a song by name"""
    q = f"track:{track_name}"
    if artist_name:
        q += f" artist:{artist_name}"
    search_results = await search(context, q, [SearchType.TRACK], 1)

    if not search_results["tracks"]["items"]:
        message = RESPONSE_MSGS["track_not_found"].format(track_name=track_name)
        if artist_name:
            message += f" by '{artist_name}'"
        raise RetryableToolError(
            message,
            additional_prompt_content=f"{message}. Try a different track name or artist name.",
            retry_after_ms=500,
        )

    track_id = search_results["tracks"]["items"][0]["id"]
    response = await start_tracks_playback_by_id(context, [track_id])

    return str(response)


# NOTE: This tool only works for Spotify Premium users
@tool(requires_auth=Spotify(scopes=["user-read-playback-state"]))
async def get_available_devices(
    context: ToolContext,
) -> Annotated[dict, "The available devices"]:
    """Get the available devices"""
    url = get_url("player_get_available_devices")
    response = await send_spotify_request(context, "GET", url)
    response.raise_for_status()
    return dict(response.json())



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import ToolExecutionError
from httpx import Response

from arcade_github.tools.issues import create_issue, create_issue_comment


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.fixture
def mock_client():
    with patch("arcade_github.tools.issues.httpx.AsyncClient") as client:
        yield client.return_value.__aenter__.return_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status_code,error_message,expected_error,func,args",
    [
        (
            422,
            "Validation Failed",
            "Error accessing.*: Validation failed",
            create_issue,
            ("owner", "repo", "title"),
        ),
        (
            401,
            "Unauthorized",
            "Error accessing.*: Failed to process request",
            create_issue_comment,
            ("owner", "repo", 1, "body"),
        ),
        (
            403,
            "API rate limit exceeded",
            "Error accessing.*: Forbidden",
            create_issue_comment,
            ("owner", "repo", 1, "body"),
        ),
        (
            401,
            "Bad credentials",
            "Error accessing.*: Failed to process request",
            create_issue,
            ("owner", "repo", "title"),
        ),
    ],
)
async def test_issue_errors(
    mock_context, mock_client, status_code, error_message, expected_error, func, args
):
    mock_client.post.return_value = Response(status_code, json={"message": error_message})

    with pytest.raises(ToolExecutionError, match=expected_error):
        await func(mock_context, *args)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "func,args,response_json,expected_assertions",
    [
        (
            create_issue,
            ("owner", "repo", "Test Issue", "This is a test issue"),
            {
                "id": 1,
                "url": "https://api.github.com/repos/owner/repo/issues/1",
                "title": "Test Issue",
                "body": "This is a test issue",
                "state": "open",
                "html_url": "https://github.com/owner/repo/issues/1",
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-01T12:00:00Z",
                "user": {"login": "testuser"},
                "assignees": [],
                "labels": [],
            },
            ["Test Issue", "https://github.com/owner/repo/issues/1"],
        ),
        (
            create_issue_comment,
            ("owner", "repo", 1, "This is a test comment"),
            {
                "id": 1,
                "url": "https://api.github.com/repos/owner/repo/issues/comments/1",
                "body": "This is a test comment",
                "user": {"login": "testuser"},
                "created_at": "2023-05-01T12:00:00Z",
                "updated_at": "2023-05-01T12:00:00Z",
            },
            ["This is a test comment", "https://api.github.com/repos/owner/repo/issues/comments/1"],
        ),
    ],
)
async def test_issue_success(
    mock_context, mock_client, func, args, response_json, expected_assertions
):
    mock_client.post.return_value = Response(201, json=response_json)

    result = await func(mock_context, *args)
    for assertion in expected_assertions:
        assert assertion in result



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import OAuth2
from arcade_tdk.errors import ToolExecutionError

from arcade_salesforce.exceptions import SalesforceToolExecutionError
from arcade_salesforce.models import SalesforceClient


# TODO: Add support for referencing an account by keywords (name, website, etc). We can use the
# get_account_data_by_keywords tool to search for accounts. If more than one is returned, we can
# raise a RetryableToolError providing the matches for the user to choose.
@tool(
    requires_auth=OAuth2(
        id="salesforce",
        scopes=["write_contact"],
    )
)
async def create_contact(
    context: ToolContext,
    account_id: Annotated[str, "The ID of the account to create the contact for."],
    last_name: Annotated[str, "The last name of the contact."],
    first_name: Annotated[str | None, "The first name of the contact."] = None,
    email: Annotated[str | None, "The email of the contact."] = None,
    phone: Annotated[str | None, "The phone number of the contact."] = None,
    mobile_phone: Annotated[str | None, "The mobile phone number of the contact."] = None,
    title: Annotated[
        str | None,
        "The title of the contact. E.g. 'CEO', 'Sales Director', 'CTO', etc.",
    ] = None,
    department: Annotated[
        str | None,
        "The department of the contact. E.g. 'Marketing', 'Sales', 'IT', etc.",
    ] = None,
    description: Annotated[str | None, "The description of the contact."] = None,
) -> Annotated[
    dict,
    "The created contact.",
]:
    """Creates a contact in Salesforce."""
    if not last_name:
        raise ToolExecutionError("Last name is required by Salesforce to create a contact.")

    client = SalesforceClient(context.get_auth_token_or_empty())
    contact = await client.create_contact(
        account_id=account_id,
        first_name=first_name,
        last_name=last_name,
        email=email,
        phone=phone,
        mobile_phone=mobile_phone,
        title=title,
        department=department,
        description=description,
    )

    if not contact.get("success"):
        raise SalesforceToolExecutionError(
            message="Failed to create contact",
            errors=contact.get("errors", []),
        )

    return {"success": True, "contactId": contact.get("id")}



from typing import Any, ClassVar
from urllib.parse import urlparse

from arcade_tdk.errors import RetryableToolError
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

MAX_ROWS_RETURNED = 1000
TEST_QUERY = "SELECT 1"


class DatabaseEngine:
    _instance: ClassVar[None] = None
    _engines: ClassVar[dict[str, AsyncEngine]] = {}

    @classmethod
    async def get_instance(cls, connection_string: str) -> AsyncEngine:
        parsed_url = urlparse(connection_string)

        # TODO: something strange with sslmode= and friends
        # query_params = parse_qs(parsed_url.query)
        # query_params = {
        #     k: v[0] for k, v in query_params.items()
        # }  # assume one value allowed for each query param

        async_connection_string = f"{parsed_url.scheme.replace('postgresql', 'postgresql+asyncpg')}://{parsed_url.netloc}{parsed_url.path}"
        key = f"{async_connection_string}"
        if key not in cls._engines:
            cls._engines[key] = create_async_engine(async_connection_string)

        # try a simple query to see if the connection is valid
        try:
            async with cls._engines[key].connect() as connection:
                await connection.execute(text(TEST_QUERY))
            return cls._engines[key]
        except Exception:
            await cls._engines[key].dispose()

            # try again
            try:
                async with cls._engines[key].connect() as connection:
                    await connection.execute(text(TEST_QUERY))
                return cls._engines[key]
            except Exception as e:
                raise RetryableToolError(
                    f"Connection failed: {e}",
                    developer_message="Connection to postgres failed.",
                    additional_prompt_content="Check the connection string and try again.",
                ) from e

    @classmethod
    async def get_engine(cls, connection_string: str) -> Any:
        engine = await cls.get_instance(connection_string)

        class ConnectionContextManager:
            def __init__(self, engine: AsyncEngine) -> None:
                self.engine = engine

            async def __aenter__(self) -> AsyncEngine:
                return self.engine

            async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                # Connection cleanup is handled by the async context manager
                pass

        return ConnectionContextManager(engine)

    @classmethod
    async def cleanup(cls) -> None:
        """Clean up all cached engines. Call this when shutting down."""
        for engine in cls._engines.values():
            await engine.dispose()
        cls._engines.clear()

    @classmethod
    def clear_cache(cls) -> None:
        """Clear the engine cache without disposing engines. Use with caution."""
        cls._engines.clear()

    @classmethod
    def sanitize_query(cls, query: str) -> str:
        """
        Sanitize a query to not break our read-only session.
        THIS IS REALLY UNSAFE AND SHOULD NOT BE USED IN PRODUCTION. USE A DATABASE CONNECTION WITH A READ-ONLY USER AND PREPARE STATEMENTS.
        There are also valid reasons for the ";" character, and this prevents that.
        """

        parts = query.split(";")
        if len(parts) > 1:
            raise RetryableToolError(
                "Multiple statements are not allowed in a single query.",
                developer_message="Multiple statements are not allowed in a single query.",
                additional_prompt_content="Split your query into multiple queries and try again.",
            )

        words = parts[0].split(" ")
        if words[0].upper().strip() != "SELECT":
            raise RetryableToolError(
                "Only SELECT queries are allowed.",
                developer_message="Only SELECT queries are allowed.",
                additional_prompt_content="Use the <DiscoverTables> and <GetTableSchema> tools to discover the tables and try again.",
            )

        return f"{query}"



import inspect
import json
import logging
from typing import Any, Callable, TypeVar

from arcade_serve.mcp.types import InitializeRequest, JSONRPCRequest, MCPMessage

logger = logging.getLogger("arcade.mcp")

T = TypeVar("T")

# Type definition for middleware functions
MessageProcessor = Callable[[Any, str], Any]


class MCPMessageProcessor:
    """
    Processes MCP messages through a chain of middleware.
    Supports both synchronous and asynchronous middleware.
    """

    def __init__(self) -> None:
        self.middleware: list[Callable[[MCPMessage, str], Any]] = []

    def add_middleware(self, mw: Callable[[MCPMessage, str], Any]) -> None:
        self.middleware.append(mw)

    async def process(self, message: Any, direction: str) -> Any:  # noqa: C901
        # First, try to parse the message if it's a string
        if isinstance(message, str):
            # Strip any whitespace including newlines
            message = message.strip()
            if not message:
                return None

            try:
                parsed = json.loads(message)
                if isinstance(parsed, dict):
                    method = parsed.get("method")
                    # Convert to appropriate message type
                    if method == "initialize" and "id" in parsed:
                        logger.debug(f"Parsed initialize request: {parsed}")
                        message = InitializeRequest(**parsed)
                    elif method and method.startswith("notifications/"):
                        # It's a notification, log it but pass through as dict
                        logger.debug(f"Received notification: {method}")
                        # Keep as parsed dict to avoid validation errors on unknown notifications
                        message = parsed
                    elif "method" in parsed and "id" in parsed:
                        # Regular method request
                        logger.debug(f"Parsed method request: {method}")
                        message = JSONRPCRequest(**parsed)
                    # Other message types can be handled similarly
            except json.JSONDecodeError:
                logger.warning(f"Failed to parse message as JSON: {message[:100]}...")
            except Exception:
                logger.exception("Error processing message")

        # Process through middleware chain
        result = message
        for mw in self.middleware:
            try:
                if inspect.iscoroutinefunction(mw):
                    result = await mw(result, direction)
                else:
                    result = mw(result, direction)
            except Exception:
                logger.exception(f"Error in middleware {mw}")
        return result

    async def process_request(self, message: Any) -> Any:
        return await self.process(message, "request")

    async def process_response(self, message: Any) -> Any:
        return await self.process(message, "response")


def create_message_processor(*middleware: MessageProcessor) -> MCPMessageProcessor:
    processor = MCPMessageProcessor()
    for m in middleware:
        if m is not None:
            processor.add_middleware(m)
    return processor



from arcade_hubspot.tools.crm.companies import get_company_data_by_keywords
from arcade_hubspot.tools.crm.contacts import create_contact, get_contact_data_by_keywords

__all__ = [
    "get_company_data_by_keywords",
    "get_contact_data_by_keywords",
    "create_contact",
]



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from firecrawl import FirecrawlApp


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def map_website(
    context: ToolContext,
    url: Annotated[str, "The base URL to start crawling from"],
    search: Annotated[str | None, "Search query to use for mapping"] = None,
    ignore_sitemap: Annotated[bool, "Ignore the website sitemap when crawling"] = True,
    include_subdomains: Annotated[bool, "Include subdomains of the website"] = False,
    limit: Annotated[int, "Maximum number of links to return"] = 5000,
) -> Annotated[dict[str, Any], "Website map data"]:
    """
    Map a website from a single URL to a map of the entire website.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    params: dict[str, Any] = {
        "ignoreSitemap": ignore_sitemap,
        "includeSubdomains": include_subdomains,
        "limit": limit,
    }
    if search:
        params["search"] = search

    map_result = app.map_url(url, params=params)

    return dict(map_result)



import decimal
import math
from decimal import Decimal
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def log(
    a: Annotated[str, "The number to take the logarithm of as a string"],
    base: Annotated[str, "The logarithmic base as a string"],
) -> Annotated[str, "The logarithm of the number with the specified base as a string"]:
    """
    Calculate the logarithm of a number with a given base
    """
    # Use Decimal for arbitrary precision
    return str(math.log(Decimal(a), Decimal(base)))


@tool
def power(
    a: Annotated[str, "The base number as a string"],
    b: Annotated[str, "The exponent as a string"],
) -> Annotated[str, "The result of raising a to the power of b as a string"]:
    """
    Calculate one number raised to the power of another
    """
    # Use Decimal for arbitrary precision
    return str(Decimal(a) ** Decimal(b))



from unittest.mock import patch

import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.fixture
def mock_httpx_client(mocker):
    with patch("arcade_dropbox.utils.httpx") as mock_httpx:
        yield mock_httpx.AsyncClient().__aenter__.return_value


@pytest.fixture
def sample_folder_entry():
    return {
        ".tag": "folder",
        "name": "test.txt",
        "path_display": "/TestFolder",
        "path_lower": "/testfolder",
        "id": "1234567890",
        "client_modified": "2025-01-01T00:00:00Z",
        "server_modified": "2025-01-01T00:00:00Z",
        "rev": "1234567890",
    }


@pytest.fixture
def sample_file_entry():
    return {
        ".tag": "file",
        "name": "test.txt",
        "path_display": "/TestFile.txt",
        "path_lower": "/testfile.txt",
        "id": "1234567890",
        "client_modified": "2025-01-01T00:00:00Z",
        "server_modified": "2025-01-01T00:00:00Z",
        "rev": "1234567890",
        "size": 1024,
    }



from unittest.mock import MagicMock, patch

import pytest
from arcade_tdk import ToolContext, ToolSecretItem
from arcade_tdk.errors import ToolExecutionError

import arcade_e2b.tools.create_chart
import arcade_e2b.tools.run_code
from arcade_e2b.enums import E2BSupportedLanguage


@pytest.fixture
def mock_run_code_sandbox():
    with patch("arcade_e2b.tools.run_code.Sandbox") as mock:
        yield mock.return_value.__enter__.return_value


@pytest.fixture
def mock_create_chart_sandbox():
    with patch("arcade_e2b.tools.create_chart.Sandbox") as mock:
        yield mock.return_value.__enter__.return_value


@pytest.fixture
def mock_context():
    return ToolContext(secrets=[ToolSecretItem(key="e2b_api_key", value="fake_api_key")])


def test_run_code_success(mock_run_code_sandbox, mock_context):
    mock_execution = MagicMock()
    mock_execution.to_json.return_value = '{"result": "success"}'
    mock_run_code_sandbox.run_code.return_value = mock_execution

    result = arcade_e2b.tools.run_code(
        mock_context, "print('Hello, World!')", E2BSupportedLanguage.PYTHON
    )
    assert result == '{"result": "success"}'


def test_run_code_error(mock_run_code_sandbox, mock_context):
    mock_execution = MagicMock()
    mock_execution.to_json.side_effect = ToolExecutionError("Execution failed")
    mock_run_code_sandbox.run_code.return_value = mock_execution

    with pytest.raises(ToolExecutionError, match="Execution failed"):
        arcade_e2b.tools.run_code(
            mock_context, "print('Hello, World!')", E2BSupportedLanguage.PYTHON
        )


def test_create_static_matplotlib_chart_success(mock_create_chart_sandbox, mock_context):
    mock_execution = MagicMock()
    mock_execution.results = [MagicMock(png="base64encodedimage")]
    mock_execution.logs.to_json.return_value = '{"logs": "log data"}'
    mock_execution.error = None
    mock_create_chart_sandbox.run_code.return_value = mock_execution

    result = arcade_e2b.tools.create_chart.create_static_matplotlib_chart(
        mock_context, "import matplotlib.pyplot as plt"
    )
    assert result == {
        "base64_image": "base64encodedimage",
        "logs": '{"logs": "log data"}',
        "error": None,
    }


def test_create_static_matplotlib_chart_error(mock_create_chart_sandbox, mock_context):
    mock_execution = MagicMock()
    mock_execution.results = []
    mock_execution.logs.to_json.return_value = '{"logs": "log data"}'
    mock_execution.error.to_json.return_value = '{"error": "some error"}'
    mock_create_chart_sandbox.run_code.return_value = mock_execution

    result = arcade_e2b.tools.create_chart.create_static_matplotlib_chart(
        mock_context, "import matplotlib.pyplot as plt"
    )
    assert result == {
        "base64_image": None,
        "logs": '{"logs": "log data"}',
        "error": '{"error": "some error"}',
    }



from arcade_microsoft.outlook_calendar.tools.create_event import create_event
from arcade_microsoft.outlook_calendar.tools.get_event import get_event
from arcade_microsoft.outlook_calendar.tools.list_events_in_time_range import (
    list_events_in_time_range,
)

__all__ = ["create_event", "get_event", "list_events_in_time_range"]



import re
from dataclasses import dataclass, field
from typing import Any

from bs4 import BeautifulSoup
from msgraph.generated.models.attendee import Attendee as GraphAttendee
from msgraph.generated.models.date_time_time_zone import DateTimeTimeZone as GraphDateTimeTimeZone
from msgraph.generated.models.email_address import EmailAddress as GraphEmailAddress
from msgraph.generated.models.event import Event as GraphEvent
from msgraph.generated.models.event_type import EventType as GraphEventType
from msgraph.generated.models.free_busy_status import FreeBusyStatus as GraphFreeBusyStatus
from msgraph.generated.models.importance import Importance as GraphImportance
from msgraph.generated.models.item_body import ItemBody as GraphItemBody
from msgraph.generated.models.location import Location as GraphLocation
from msgraph.generated.models.recipient import Recipient as GraphRecipient
from msgraph.generated.models.response_status import ResponseStatus as GraphResponseStatus
from msgraph.generated.models.response_type import ResponseType as GraphResponseType


@dataclass
class Attendee:
    """An attendee of a calendar event."""

    name: str = ""
    address: str = ""
    response: str = ""

    @classmethod
    def from_sdk(cls, attendee: GraphAttendee) -> "Attendee":
        """Convert a Microsoft Graph SDK Attendee object to an Attendee dataclass."""
        return cls(
            name=attendee.email_address.name
            if attendee.email_address and attendee.email_address.name
            else "",
            address=attendee.email_address.address
            if attendee.email_address and attendee.email_address.address
            else "",
            response=attendee.status.response
            if attendee.status and attendee.status.response
            else "",
        )

    def to_dict(self) -> dict[str, str]:
        return {
            "name": self.name,
            "address": self.address,
            "response": self.response,
        }

    def to_sdk(self) -> GraphAttendee:
        """Convert an Attendee dataclass to a Microsoft Graph SDK Attendee object."""
        return GraphAttendee(
            email_address=GraphEmailAddress(name=self.name, address=self.address),
            status=GraphResponseStatus(
                response=GraphResponseType(self.response)
                if self.response
                else GraphResponseType.None_
            ),
        )


@dataclass
class Organizer:
    """The organizer of an event."""

    name: str = ""
    address: str = ""

    @classmethod
    def from_sdk(cls, organizer: GraphRecipient) -> "Organizer":
        """Convert a Microsoft Graph SDK Organizer object to an Organizer dataclass."""
        return cls(
            name=organizer.email_address.name
            if organizer.email_address and organizer.email_address.name
            else "",
            address=organizer.email_address.address
            if organizer.email_address and organizer.email_address.address
            else "",
        )

    def to_dict(self) -> dict[str, str]:
        return {
            "name": self.name,
            "address": self.address,
        }

    def to_sdk(self) -> GraphRecipient:
        """Convert an Organizer dataclass to a Microsoft Graph SDK Organizer object."""
        recipient = GraphRecipient(
            email_address=GraphEmailAddress(name=self.name, address=self.address)
        )
        return recipient


@dataclass
class DateTimeTimeZone:
    """Time information for an event."""

    date_time: str = ""
    time_zone: str = ""

    @classmethod
    def from_sdk(cls, date_time_time_zone: GraphDateTimeTimeZone) -> "DateTimeTimeZone":
        """Convert a Microsoft Graph SDK DateTimeTimeZone object to a TimeInfo dataclass."""
        return cls(
            date_time=date_time_time_zone.date_time or "",
            time_zone=date_time_time_zone.time_zone or "",
        )

    def to_dict(self) -> dict[str, str]:
        return {
            "dateTime": self.date_time,
            "timeZone": self.time_zone,
        }

    def to_sdk(self) -> GraphDateTimeTimeZone:
        """Convert a TimeInfo dataclass to a Microsoft Graph SDK DateTimeTimeZone object."""
        return GraphDateTimeTimeZone(date_time=self.date_time, time_zone=self.time_zone)


@dataclass
class ResponseStatus:
    """The response status for an event."""

    response: str = ""

    @classmethod
    def from_sdk(cls, response_status: GraphResponseStatus) -> "ResponseStatus":
        """Convert a Microsoft Graph SDK ResponseStatus object to a ResponseStatus dataclass."""
        response_value = (
            str(response_status.response.value)
            if response_status.response and hasattr(response_status.response, "value")
            else ""
        )
        return cls(response=response_value)

    def to_dict(self) -> dict[str, str]:
        return {
            "response": self.response,
        }

    def to_sdk(self) -> GraphResponseStatus:
        """Convert a ResponseStatus dataclass to a Microsoft Graph SDK ResponseStatus object."""
        return GraphResponseStatus(response=GraphResponseType(self.response))


@dataclass
class Event:
    """A calendar event in Outlook."""

    attendees: list[Attendee] = field(default_factory=list)
    body: str = ""
    end: DateTimeTimeZone | None = None
    has_attachments: bool = False
    importance: str = ""
    is_all_day: bool = False
    is_cancelled: bool = False
    is_draft: bool = False
    is_online_meeting: bool = False
    is_organizer: bool = False
    location: str = ""
    online_meeting_url: str = ""
    organizer: Organizer | None = None
    id: str = ""
    response_status: ResponseStatus | None = None
    show_as: str = ""
    start: DateTimeTimeZone | None = None
    subject: str = ""
    type: str = ""
    web_link: str = ""
    event_id: str = ""  # The unique identifier of the event. Read-only.

    @staticmethod
    def _safe_str(value: Any) -> str:
        if not value:
            return ""
        if isinstance(value, bytes | bytearray):
            return value.decode("utf-8", errors="ignore")
        return str(value)

    @staticmethod
    def _safe_bool(value: Any) -> bool:
        return bool(value)

    @staticmethod
    def _parse_body(mime: str) -> str:
        if not mime:
            return ""
        soup = BeautifulSoup(mime, "html.parser")
        text = soup.get_text(separator=" ")
        # Replace multiple newlines with a single newline
        text = re.sub(r"\n+", "\n", text)
        # Replace multiple spaces with a single space
        text = re.sub(r"\s+", " ", text)
        # Replace sequences of dots (likely from horizontal lines) with a single newline
        text = re.sub(r"\.{3,}", "\n---\n", text)
        # Remove leading/trailing whitespace from each line
        text = "\n".join(line.strip() for line in text.split("\n"))
        return text

    @classmethod
    def from_sdk(cls, event: GraphEvent) -> "Event":
        """Convert a Microsoft Graph SDK Event object to an Event dataclass."""
        body_mime = event.body.content if event.body and event.body.content else ""
        body = cls._parse_body(body_mime)

        attendees = [Attendee.from_sdk(a) for a in event.attendees if a] if event.attendees else []
        start = DateTimeTimeZone.from_sdk(event.start) if event.start else None
        end = DateTimeTimeZone.from_sdk(event.end) if event.end else None
        organizer = Organizer.from_sdk(event.organizer) if event.organizer else None
        response_status = (
            ResponseStatus.from_sdk(event.response_status) if event.response_status else None
        )

        return cls(
            attendees=attendees,
            body=body,
            end=end,
            has_attachments=cls._safe_bool(event.has_attachments),
            importance=cls._safe_str(str(event.importance.value)) if event.importance else "",
            is_all_day=cls._safe_bool(event.is_all_day),
            is_cancelled=cls._safe_bool(event.is_cancelled),
            is_draft=cls._safe_bool(event.is_draft),
            is_online_meeting=cls._safe_bool(event.is_online_meeting),
            is_organizer=cls._safe_bool(event.is_organizer),
            location=cls._safe_str(event.location.display_name if event.location else ""),
            online_meeting_url=cls._safe_str(event.online_meeting_url),
            organizer=organizer,
            id=cls._safe_str(event.id),
            response_status=response_status,
            show_as=cls._safe_str(str(event.show_as.value)) if event.show_as else "",
            start=start,
            subject=cls._safe_str(event.subject),
            type=cls._safe_str(str(event.type.value)) if event.type else "",
            web_link=cls._safe_str(event.web_link),
            event_id=cls._safe_str(event.id),
        )

    def to_dict(self) -> dict[str, Any]:
        """Converts the Event dataclass to a dictionary."""
        return {
            "attendees": [attendee.to_dict() for attendee in self.attendees],
            "body": self.body,
            "end": self.end.to_dict() if self.end else None,
            "has_attachments": self.has_attachments,
            "importance": self.importance,
            "is_all_day": self.is_all_day,
            "is_cancelled": self.is_cancelled,
            "is_draft": self.is_draft,
            "is_online_meeting": self.is_online_meeting,
            "is_organizer": self.is_organizer,
            "location": self.location,
            "online_meeting_url": self.online_meeting_url,
            "organizer": self.organizer.to_dict() if self.organizer else None,
            "id": self.id,
            "response_status": self.response_status.to_dict() if self.response_status else None,
            "show_as": self.show_as,
            "start": self.start.to_dict() if self.start else None,
            "subject": self.subject,
            "type": self.type,
            "web_link": self.web_link,
            "event_id": self.event_id,
        }

    def to_sdk(self) -> GraphEvent:
        """Convert an Event dataclass to a Microsoft Graph SDK Event object."""
        return GraphEvent(
            attendees=[attendee.to_sdk() for attendee in self.attendees],
            body=GraphItemBody(content=self.body),
            end=self.end.to_sdk() if self.end else None,
            has_attachments=self.has_attachments,
            importance=GraphImportance(self.importance) if self.importance else None,
            is_all_day=self.is_all_day,
            is_cancelled=self.is_cancelled,
            is_draft=self.is_draft,
            is_online_meeting=self.is_online_meeting,
            is_organizer=self.is_organizer,
            location=GraphLocation(display_name=self.location),
            online_meeting_url=self.online_meeting_url,
            organizer=self.organizer.to_sdk() if self.organizer else None,
            id=self.id,
            response_status=self.response_status.to_sdk() if self.response_status else None,
            show_as=GraphFreeBusyStatus(self.show_as) if self.show_as else None,
            start=self.start.to_sdk() if self.start else None,
            subject=self.subject,
            type=GraphEventType(self.type) if self.type else None,
            web_link=self.web_link,
        )



import contextlib
import re
from datetime import datetime
from typing import Any, cast
from zoneinfo import ZoneInfo

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from serpapi import Client as SerpClient

from arcade_google_maps.constants import (
    DEFAULT_GOOGLE_MAPS_COUNTRY,
    DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
    DEFAULT_GOOGLE_MAPS_LANGUAGE,
    DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
)
from arcade_google_maps.enums import GoogleMapsDistanceUnit, GoogleMapsTravelMode
from arcade_google_maps.exceptions import CountryNotFoundError, LanguageNotFoundError
from arcade_google_maps.google_data import COUNTRY_CODES, LANGUAGE_CODES


def prepare_params(engine: str, **kwargs: Any) -> dict[str, Any]:
    """
    Prepares a parameters dictionary for the SerpAPI call.

    Parameters:
        engine: The engine name (e.g., "google", "google_finance").
        kwargs: Any additional parameters to include.

    Returns:
        A dictionary containing the base parameters plus any extras,
        excluding any parameters whose value is None.
    """
    params = {"engine": engine}
    params.update({k: v for k, v in kwargs.items() if v is not None})
    return params


def call_serpapi(context: ToolContext, params: dict) -> dict:
    """
    Execute a search query using the SerpAPI client and return the results as a dictionary.

    Args:
        context: The tool context containing required secrets.
        params: A dictionary of parameters for the SerpAPI search.

    Returns:
        The search results as a dictionary.
    """
    api_key = context.get_secret("SERP_API_KEY")
    client = SerpClient(api_key=api_key)
    try:
        search = client.search(params)
        return cast(dict[str, Any], search.as_dict())
    except Exception as e:
        # SerpAPI error messages sometimes contain the API key, so we need to sanitize it
        sanitized_e = re.sub(r"(api_key=)[^ &]+", r"\1***", str(e))
        raise ToolExecutionError(
            message="Failed to fetch search results",
            developer_message=sanitized_e,
        )


def get_google_maps_directions(
    context: ToolContext,
    origin_address: str | None = None,
    destination_address: str | None = None,
    origin_latitude: str | None = None,
    origin_longitude: str | None = None,
    destination_latitude: str | None = None,
    destination_longitude: str | None = None,
    language: str | None = DEFAULT_GOOGLE_MAPS_LANGUAGE,
    country: str | None = DEFAULT_GOOGLE_MAPS_COUNTRY,
    distance_unit: GoogleMapsDistanceUnit = DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
    travel_mode: GoogleMapsTravelMode = DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
) -> list[dict[str, Any]]:
    """Get directions from Google Maps.

    Provide either all(origin_address, destination_address) or
    all(origin_latitude, origin_longitude, destination_latitude, destination_longitude).

    Args:
        context: Tool context containing required Serp API Key secret.
        origin_address: Origin address.
        destination_address: Destination address.
        origin_latitude: Origin latitude.
        origin_longitude: Origin longitude.
        destination_latitude: Destination latitude.
        destination_longitude: Destination longitude.
        language: Language to use in the Google Maps search. Defaults to 'en' (English).
        country: 2-letter country code to use in the Google Maps search. Defaults to None
            (no country is specified).
        distance_unit: Distance unit to use in the Google Maps search. Defaults to 'km'
            (kilometers).
        travel_mode: Travel mode to use in the Google Maps search. Defaults to 'best'
            (best mode).

    Returns:
        The directions from Google Maps.
    """
    if isinstance(language, str):
        language = language.lower()

    if language not in LANGUAGE_CODES:
        raise LanguageNotFoundError(language)

    params = prepare_params(
        engine="google_maps_directions",
        hl=language,
        distance_unit=distance_unit.to_api_value(),
        travel_mode=travel_mode.to_api_value(),
    )

    if any([
        origin_latitude,
        origin_longitude,
        destination_latitude,
        destination_longitude,
    ]) and any([origin_address, destination_address]):
        raise ValueError("Either coordinates or addresses must be provided, not both")

    elif all([origin_latitude, origin_longitude, destination_latitude, destination_longitude]):
        params["start_coords"] = f"{origin_latitude},{origin_longitude}"
        params["end_coords"] = f"{destination_latitude},{destination_longitude}"

    elif all([origin_address, destination_address]):
        params["start_addr"] = str(origin_address)
        params["end_addr"] = str(destination_address)

    else:
        raise ValueError("Either coordinates or addresses must be provided")

    if country:
        country = country.lower()
        if country not in COUNTRY_CODES:
            raise CountryNotFoundError(country)
        params["gl"] = country

    results = call_serpapi(context, params)

    directions = cast(list[dict[str, Any]], results.get("directions", []))

    for direction in directions:
        clean_google_maps_direction(direction)

        if "arrive_around" in direction:
            direction["arrive_around"] = enrich_google_maps_arrive_around(
                direction["arrive_around"]
            )

    return directions


def clean_google_maps_direction(direction: dict[str, Any]) -> None:
    for trip in direction.get("trips", []):
        with contextlib.suppress(KeyError):
            del trip["start_stop"]["data_id"]
            del trip["end_stop"]["data_id"]

        for detail in trip.get("details", []):
            with contextlib.suppress(KeyError):
                del detail["geo_photo"]
                del detail["gps_coordinates"]

        for stop in trip.get("stops", []):
            with contextlib.suppress(KeyError):
                del stop["data_id"]


def enrich_google_maps_arrive_around(timestamp: int | None) -> dict[str, Any]:
    if not timestamp:
        return {}

    dt = datetime.fromtimestamp(timestamp, tz=ZoneInfo("UTC")).isoformat()
    return {"datetime": dt, "timestamp": timestamp}



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_asana
from arcade_asana.tools import (
    create_task,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_asana)


@tool_eval()
def create_task_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="create task eval suite",
        system_message="You are an AI assistant with access to Asana tools. Use them to help the user with their tasks.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Create task with name, description, start and due dates",
        user_message="Create a task with the name 'Hello World' and the description 'This is a task description' starting on 2025-05-05 and due on 2025-05-11.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_task,
                args={
                    "name": "Hello World",
                    "description": "This is a task description",
                    "start_date": "2025-05-05",
                    "due_date": "2025-05-11",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=1 / 4),
            BinaryCritic(critic_field="description", weight=1 / 4),
            BinaryCritic(critic_field="start_date", weight=1 / 4),
            BinaryCritic(critic_field="due_date", weight=1 / 4),
        ],
    )

    suite.add_case(
        name="Create task with name and tag names",
        user_message="Create a task with the name 'Hello World' and the tags 'My Tag' and 'My Other Tag'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_task,
                args={
                    "name": "Hello World",
                    "tags": ["My Tag", "My Other Tag"],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=0.5),
            BinaryCritic(critic_field="tags", weight=0.5),
        ],
    )

    suite.add_case(
        name="Create task with name and tag IDs",
        user_message="Create a task with the name 'Hello World' and the tags '1234567890' and '1234567891'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_task,
                args={
                    "name": "Hello World",
                    "tags": ["1234567890", "1234567891"],
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="name", weight=0.5),
            BinaryCritic(critic_field="tags", weight=0.5),
        ],
    )

    return suite



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_microsoft.outlook_calendar import get_event
from evals.outlook_calendar.additional_messages import get_event_additional_messages

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_tool(get_event, "Microsoft")


@tool_eval()
def outlook_calendar_get_event_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Outlook Calendar get event tool."""
    suite = EvalSuite(
        name="Outlook Calendar Get Event Evaluation",
        system_message=(
            "You are an AI that has access to tools to view and manage calendar events. "
            "The current time date and time is April 25, 2025, 5:18 PM PST."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get event by id after listing events",
        user_message="tell me more about the first event",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_event,
                args={
                    "event_id": "AAMkADAwATM0MDAAMi04Y2Y1LTQ3MTEALTAwAi0wMAoARgAAAyXxSd3UxTpCkDpGouEg0JMBAFuxokOLZRtDncM4",  # noqa: E501
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="event_id", weight=1.0),
        ],
        additional_messages=get_event_additional_messages,
    )

    return suite



from collections.abc import Callable
from unittest.mock import patch

import pytest
from arcade_tdk import ToolContext

from arcade_jira.exceptions import JiraToolExecutionError, MultipleItemsFoundError, NotFoundError
from arcade_jira.utils import clean_priority_dict, find_unique_priority


@pytest.mark.asyncio
async def test_find_unique_priority_by_id_success(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_priority_dict: Callable,
):
    sample_priority = build_priority_dict()
    priority_response = mock_httpx_response(200, sample_priority)
    mock_httpx_client.get.return_value = priority_response

    response = await find_unique_priority(mock_context, sample_priority["id"], "123")
    assert response == clean_priority_dict(sample_priority)


@pytest.mark.asyncio
@patch("arcade_jira.tools.priorities.find_priorities_by_project")
async def test_find_unique_priority_by_name_with_a_single_match(
    mock_find_priorities_by_project,
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_priority_dict: Callable,
):
    sample_project = build_project_dict()
    sample_priority = build_priority_dict()

    # It will first try to get the priority by ID
    get_priority_by_id_response = mock_httpx_response(404, {})

    # When it tries to get the priority by name, it will first query the project data
    get_project_by_id_response = mock_httpx_response(200, sample_project)

    # Then it will query the priorities available to the project
    mock_find_priorities_by_project.return_value = {
        "project": sample_project,
        "priorities_available": [sample_priority],
    }

    mock_httpx_client.get.side_effect = [
        get_priority_by_id_response,
        get_project_by_id_response,
    ]

    response = await find_unique_priority(
        mock_context, sample_priority["name"].lower(), sample_project["id"]
    )
    assert response == clean_priority_dict(sample_priority)


@pytest.mark.asyncio
@patch("arcade_jira.tools.priorities.find_priorities_by_project")
async def test_find_unique_priority_by_name_when_project_does_not_exist(
    mock_find_priorities_by_project,
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_priority_dict: Callable,
    build_project_search_response_dict: Callable,
):
    sample_project = build_project_dict()
    sample_priority = build_priority_dict()

    # It will first try to get the priority by ID
    get_priority_by_id_response = mock_httpx_response(404, {})

    # When it tries to get the project by id, we'll simulate a 404 error
    get_project_by_id_response = mock_httpx_response(404, {})

    # And also simulate no results found from search_projects
    search_projects_response = mock_httpx_response(200, build_project_search_response_dict([]))

    # We'll still simulate a find_priorities_by_project response, but this should not be called
    mock_find_priorities_by_project.return_value = {
        "project": sample_project,
        "priorities_available": [sample_priority],
    }

    mock_httpx_client.get.side_effect = [
        get_priority_by_id_response,
        get_project_by_id_response,
        search_projects_response,
    ]

    with pytest.raises(JiraToolExecutionError) as exc:
        await find_unique_priority(
            mock_context, sample_priority["name"].lower(), sample_project["id"]
        )

    mock_find_priorities_by_project.assert_not_called()
    assert f"Project not found with name/key/ID '{sample_project['id']}'" in exc.value.message


@pytest.mark.asyncio
@patch("arcade_jira.tools.priorities.find_priorities_by_project")
async def test_find_unique_priority_by_name_with_multiple_priorities_but_zero_matches(
    mock_find_priorities_by_project,
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_priority_dict: Callable,
):
    sample_project = build_project_dict()

    sample_priority = build_priority_dict()
    other_priority1 = build_priority_dict(name=sample_priority["name"] + "1")
    other_priority2 = build_priority_dict(name=sample_priority["name"] + "2")

    # It will first try to get the priority by ID
    get_priority_by_id_response = mock_httpx_response(404, {})

    # When it tries to get the priority by name, it will first query the project data
    get_project_by_id_response = mock_httpx_response(200, sample_project)

    # Then it will query the priorities available to the project
    mock_find_priorities_by_project.return_value = {
        "project": sample_project,
        "priorities_available": [other_priority1, other_priority2],
    }

    mock_httpx_client.get.side_effect = [
        get_priority_by_id_response,
        get_project_by_id_response,
    ]

    with pytest.raises(NotFoundError) as exc:
        await find_unique_priority(
            mock_context, sample_priority["name"].lower(), sample_project["id"]
        )

    assert (
        f"Priority not found with ID or name '{sample_priority['name'].lower()}'"
        == exc.value.message
    )


@pytest.mark.asyncio
@patch("arcade_jira.tools.priorities.find_priorities_by_project")
async def test_find_unique_priority_by_name_with_multiple_priorities_but_one_match(
    mock_find_priorities_by_project,
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_priority_dict: Callable,
):
    sample_project = build_project_dict()
    sample_priority = build_priority_dict()
    other_priority1 = build_priority_dict()
    other_priority2 = build_priority_dict()

    # It will first try to get the priority by ID
    get_priority_by_id_response = mock_httpx_response(404, {})

    # When it tries to get the priority by name, it will first query the project data
    get_project_by_id_response = mock_httpx_response(200, sample_project)

    # Then it will query the priorities available to the project
    mock_find_priorities_by_project.return_value = {
        "project": sample_project,
        "priorities_available": [sample_priority, other_priority1, other_priority2],
    }

    mock_httpx_client.get.side_effect = [
        get_priority_by_id_response,
        get_project_by_id_response,
    ]

    response = await find_unique_priority(
        mock_context, sample_priority["name"].lower(), sample_project["id"]
    )
    assert response == clean_priority_dict(sample_priority)


@pytest.mark.asyncio
@patch("arcade_jira.tools.priorities.find_priorities_by_project")
async def test_find_unique_priority_by_name_with_multiple_priorities_and_multiple_matches(
    mock_find_priorities_by_project,
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_priority_dict: Callable,
):
    sample_project = build_project_dict()
    sample_priority = build_priority_dict()
    other_priority1 = build_priority_dict(name=sample_priority["name"])
    other_priority2 = build_priority_dict()

    # It will first try to get the priority by ID
    get_priority_by_id_response = mock_httpx_response(404, {})

    # When it tries to get the priority by name, it will first query the project data
    get_project_by_id_response = mock_httpx_response(200, sample_project)

    # Then it will query the priorities available to the project
    mock_find_priorities_by_project.return_value = {
        "project": sample_project,
        "priorities_available": [sample_priority, other_priority1, other_priority2],
    }

    mock_httpx_client.get.side_effect = [
        get_priority_by_id_response,
        get_project_by_id_response,
    ]

    with pytest.raises(MultipleItemsFoundError) as exc:
        await find_unique_priority(
            mock_context, sample_priority["name"].lower(), sample_project["id"]
        )

    assert sample_priority["id"] in exc.value.message
    assert other_priority1["id"] in exc.value.message
    assert other_priority2["id"] not in exc.value.message



from zoneinfo import available_timezones

from arcade_tdk.errors import RetryableToolError


class GoogleToolError(Exception):
    """Base exception for Google tool errors."""

    def __init__(self, message: str, developer_message: str | None = None):
        self.message = message
        self.developer_message = developer_message
        super().__init__(self.message)

    def __str__(self) -> str:
        base_message = self.message
        if self.developer_message:
            return f"{base_message} (Developer: {self.developer_message})"
        return base_message


class RetryableGoogleToolError(RetryableToolError):
    """Raised when there's an error in a Google tool that can be retried."""

    pass


class GoogleServiceError(GoogleToolError):
    """Raised when there's an error building or using the Google service."""

    pass


class GmailToolError(GoogleToolError):
    """Raised when there's an error in the Gmail tools."""

    pass


class GoogleCalendarToolError(GoogleToolError):
    """Raised when there's an error in the Google Calendar tools."""

    pass


class InvalidTimezoneError(RetryableGoogleToolError):
    """Raised when a timezone is provided that is not supported by Python's zoneinfo."""

    def __init__(self, timezone_str: str):
        self.timezone_str = timezone_str
        available_timezones_msg = (
            "Here is a list of valid timezones (from Python's zoneinfo.available_timezones()): "
            f"{available_timezones()}"
        )
        super().__init__(
            f"Invalid timezone: '{timezone_str}'",
            developer_message=available_timezones_msg,
            additional_prompt_content=available_timezones_msg,
        )


class GoogleDriveToolError(GoogleToolError):
    """Raised when there's an error in the Google Drive tools."""

    pass


class GoogleDocsToolError(GoogleToolError):
    """Raised when there's an error in the Google Docs tools."""

    pass



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

from arcade_confluence.client import ConfluenceClientV2
from arcade_confluence.enums import AttachmentSortOrder
from arcade_confluence.utils import remove_none_values


@tool(
    requires_auth=Atlassian(
        scopes=["read:attachment:confluence"],
    )
)
async def list_attachments(
    context: ToolContext,
    sort_order: Annotated[
        AttachmentSortOrder,
        "The order of the attachments to sort by. Defaults to created-date-newest-to-oldest",
    ] = AttachmentSortOrder.CREATED_DATE_DESCENDING,
    limit: Annotated[
        int, "The maximum number of attachments to return. Defaults to 25. Max is 250"
    ] = 25,
    pagination_token: Annotated[
        str | None,
        "The pagination token to use for the next page of results",
    ] = None,
) -> Annotated[dict, "The attachments"]:
    """List attachments in a workspace"""
    client = ConfluenceClientV2(context.get_auth_token_or_empty())
    params = remove_none_values({
        "sort": sort_order.to_api_value(),
        "limit": max(1, min(limit, 250)),
        "cursor": pagination_token,
    })
    attachments = await client.get("attachments", params=params)
    return client.transform_get_attachments_response(attachments)


@tool(
    requires_auth=Atlassian(
        scopes=["read:attachment:confluence"],
    )
)
async def get_attachments_for_page(
    context: ToolContext,
    page_identifier: Annotated[str, "The ID or title of the page to get attachments for"],
    limit: Annotated[
        int, "The maximum number of attachments to return. Defaults to 25. Max is 250"
    ] = 25,
    pagination_token: Annotated[
        str | None,
        "The pagination token to use for the next page of results",
    ] = None,
) -> Annotated[dict, "The attachments"]:
    """Get attachments for a page by its ID or title.

    If a page title is provided, then the first page with an exact matching title will be returned.
    """
    client = ConfluenceClientV2(context.get_auth_token_or_empty())
    page_id = await client.get_page_id(page_identifier)

    params = remove_none_values({
        "limit": max(1, min(limit, 250)),
        "cursor": pagination_token,
    })
    attachments = await client.get(f"pages/{page_id}/attachments", params=params)
    return client.transform_get_attachments_response(attachments)



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Google
from arcade_tdk.errors import RetryableToolError

from arcade_google_sheets.models import (
    SheetDataInput,
    Spreadsheet,
    SpreadsheetProperties,
)
from arcade_google_sheets.utils import (
    build_sheets_service,
    create_sheet,
    parse_write_to_cell_response,
    validate_write_to_cell_params,
)


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    )
)
def create_spreadsheet(
    context: ToolContext,
    title: Annotated[str, "The title of the new spreadsheet"] = "Untitled spreadsheet",
    data: Annotated[
        str | None,
        "The data to write to the spreadsheet. A JSON string "
        "(property names enclosed in double quotes) representing a dictionary that "
        "maps row numbers to dictionaries that map column letters to cell values. "
        "For example, data[23]['C'] would be the value of the cell in row 23, column C. "
        "Type hint: dict[int, dict[str, Union[int, float, str, bool]]]",
    ] = None,
) -> Annotated[dict, "The created spreadsheet's id and title"]:
    """Create a new spreadsheet with the provided title and data in its first sheet

    Returns the newly created spreadsheet's id and title
    """
    service = build_sheets_service(context.get_auth_token_or_empty())

    try:
        sheet_data = SheetDataInput(data=data)  # type: ignore[arg-type]
    except Exception as e:
        msg = "Invalid JSON or unexpected data format for parameter `data`"
        raise RetryableToolError(
            message=msg,
            additional_prompt_content=f"{msg}: {e}",
            retry_after_ms=100,
        )

    spreadsheet = Spreadsheet(
        properties=SpreadsheetProperties(title=title),
        sheets=[create_sheet(sheet_data)],
    )

    body = spreadsheet.model_dump()

    response = (
        service.spreadsheets()
        .create(body=body, fields="spreadsheetId,spreadsheetUrl,properties/title")
        .execute()
    )

    return {
        "title": response["properties"]["title"],
        "spreadsheetId": response["spreadsheetId"],
        "spreadsheetUrl": response["spreadsheetUrl"],
    }


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/drive.file"],
    )
)
def write_to_cell(
    context: ToolContext,
    spreadsheet_id: Annotated[str, "The id of the spreadsheet to write to"],
    column: Annotated[str, "The column string to write to. For example, 'A', 'F', or 'AZ'"],
    row: Annotated[int, "The row number to write to"],
    value: Annotated[str, "The value to write to the cell"],
    sheet_name: Annotated[
        str, "The name of the sheet to write to. Defaults to 'Sheet1'"
    ] = "Sheet1",
) -> Annotated[dict, "The status of the operation"]:
    """
    Write a value to a single cell in a spreadsheet.
    """
    service = build_sheets_service(context.get_auth_token_or_empty())
    validate_write_to_cell_params(service, spreadsheet_id, sheet_name, column, row)

    range_ = f"'{sheet_name}'!{column.upper()}{row}"
    body = {
        "range": range_,
        "majorDimension": "ROWS",
        "values": [[value]],
    }

    sheet_properties = (
        service.spreadsheets()
        .values()
        .update(
            spreadsheetId=spreadsheet_id,
            range=range_,
            valueInputOption="USER_ENTERED",
            includeValuesInResponse=True,
            body=body,
        )
        .execute()
    )

    return parse_write_to_cell_response(sheet_properties)



import pytest


@pytest.fixture
def sample_drive_file_tree_request_responses() -> tuple[dict, list]:
    files_list = {
        "files": [
            # Shared Drive 1 files and folders
            {
                "id": "19WVyQndQsc0AxxfdrIt5CvDQd6r-BvpqnB8bWZoL7Xk",
                "name": "shared-1-folder-1-doc-1",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["1dCOCdPxhTqiB3j3bWrIWM692ZbL8dyjt"],
                "createdTime": "2025-02-26T00:28:20.571Z",
                "modifiedTime": "2025-02-26T00:28:30.773Z",
                "driveId": "0AFqcR6obkydtUk9PVA",
                "size": "1024",
            },
            {
                "id": "1dCOCdPxhTqiB3j3bWrIWM692ZbL8dyjt",
                "name": "shared-1-folder-1",
                "mimeType": "application/vnd.google-apps.folder",
                "parents": ["0AFqcR6obkydtUk9PVA"],
                "createdTime": "2025-02-26T00:27:45.526Z",
                "modifiedTime": "2025-02-26T00:27:45.526Z",
                "driveId": "0AFqcR6obkydtUk9PVA",
            },
            {
                "id": "1didt_h-tDjuJ-dmYtHUSyOCPci30K_kSszvg0G3tKBM",
                "name": "shared-1-doc-1",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["0AFqcR6obkydtUk9PVA"],
                "createdTime": "2025-02-26T00:27:19.287Z",
                "modifiedTime": "2025-02-26T00:27:26.079Z",
                "driveId": "0AFqcR6obkydtUk9PVA",
                "size": "1024",
            },
            # My Drive files and folders
            {
                "id": "1vB6sv0MD0hYSraYvWU_fcci3GN_-Jf4g-LfyXdG8ZMo",
                "name": "The Birth of MX Engineering",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["0AIbBwO2hjeHqUk9PVA"],
                "createdTime": "2025-01-24T06:34:22.305Z",
                "modifiedTime": "2025-02-25T21:54:30.632Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
                "size": "6634",
            },
            {
                "id": "1wv2dmYo0skJTI59ZIcwH9vm-wt7psMwXTvihuEGeHeI",
                "name": "test document 1.1.1",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["1J92V9yvVWm_uNHq3CCY4wyG1H9B6iiwO"],
                "createdTime": "2025-02-25T17:59:03.325Z",
                "modifiedTime": "2025-02-25T17:59:11.445Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
                "size": "1024",
            },
            {
                "id": "1J92V9yvVWm_uNHq3CCY4wyG1H9B6iiwO",
                "name": "test folder 1.1",
                "mimeType": "application/vnd.google-apps.folder",
                "parents": ["1gqioaHG53jPVeJN5gBpHoO-GWtwiJcLo"],
                "createdTime": "2025-02-25T17:58:58.987Z",
                "modifiedTime": "2025-02-25T17:58:58.987Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
            },
            {
                "id": "1DSmL7d07kjT6b6L-t4JIT06ElUbZ1q0K6_gEpn_UGZ8",
                "name": "test document 1.2",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["1gqioaHG53jPVeJN5gBpHoO-GWtwiJcLo"],
                "createdTime": "2025-02-25T17:58:38.628Z",
                "modifiedTime": "2025-02-25T17:58:46.713Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
                "size": "1024",
            },
            {
                "id": "1Fcxz7HsyO2Zyc-5DTD3zBQnaVrZwD29BP9KD9rPnYfE",
                "name": "test document 1.1",
                "mimeType": "application/vnd.google-apps.document",
                "parents": ["1gqioaHG53jPVeJN5gBpHoO-GWtwiJcLo"],
                "createdTime": "2025-02-25T17:57:53.850Z",
                "modifiedTime": "2025-02-25T17:58:28.745Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
                "size": "1024",
            },
            {
                "id": "1gqioaHG53jPVeJN5gBpHoO-GWtwiJcLo",
                "name": "test folder 1",
                "mimeType": "application/vnd.google-apps.folder",
                "parents": ["0AIbBwO2hjeHqUk9PVA"],
                "createdTime": "2025-02-25T17:57:46.036Z",
                "modifiedTime": "2025-02-25T17:57:46.036Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "one_new_tool_everyday",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": True,
                        "permissionId": "00356981722324419750",
                        "emailAddress": "one_new_tool_everyday@arcade.dev",
                    }
                ],
            },
            {
                "id": "16PUe97yGQeOjQgrgd54iCoxzid4SEvu_J33P_ELd5r8",
                "name": "Hello world presentation",
                "mimeType": "application/vnd.google-apps.presentation",
                "createdTime": "2025-02-18T20:48:52.786Z",
                "modifiedTime": "2025-02-19T23:31:20.483Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "john.doe",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": False,
                        "permissionId": "06420661154928749996",
                        "emailAddress": "john.doe@arcade.dev",
                    }
                ],
                "size": "15774558",
            },
            {
                "id": "1nG7lSvIyK05N9METPczVJa4iGgE7uoo-A6zpqjpUsDY",
                "name": "Shared doc 1",
                "mimeType": "application/vnd.google-apps.document",
                "createdTime": "2025-02-19T18:51:44.622Z",
                "modifiedTime": "2025-02-19T19:30:39.773Z",
                "owners": [
                    {
                        "kind": "drive#user",
                        "displayName": "theboss",
                        "photoLink": "https://lh3.googleusercontent.com/a-/photo.png",
                        "me": False,
                        "permissionId": "11571864250637401873",
                        "emailAddress": "theboss@arcade.dev",
                    }
                ],
                "size": "2700",
            },
        ],
    }

    drives_get = [
        {
            "id": "0AFqcR6obkydtUk9PVA",
            "name": "Shared Drive 1",
        }
    ]

    return files_list, drives_get



from arcade_firecrawl.tools.crawl import (
    cancel_crawl,
    crawl_website,
    get_crawl_data,
    get_crawl_status,
)
from arcade_firecrawl.tools.map import map_website
from arcade_firecrawl.tools.scrape import scrape_url

__all__ = [
    "cancel_crawl",
    "crawl_website",
    "get_crawl_data",
    "get_crawl_status",
    "map_website",
    "scrape_url",
]



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_slack
from arcade_slack.models import ConversationType
from arcade_slack.tools.chat import list_conversations

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
# Register the Slack tools
catalog.add_module(arcade_slack)


@tool_eval()
def list_conversations_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools listing conversations."""
    suite = EvalSuite(
        name="Slack Messaging Tools Evaluation",
        system_message="You are an AI assistant that can interact with Slack to send messages and get information from conversations, users, etc.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List all conversations I am a member of",
        user_message="List all conversations I am a member of",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List 10 conversations I am a member of",
        user_message="List 10 conversations I am a member of",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": None,
                    "limit": 10,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=0.5),
            BinaryCritic(critic_field="limit", weight=0.5),
        ],
    )

    suite.add_case(
        name="List all public channels",
        user_message="List all public channels",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [ConversationType.PUBLIC_CHANNEL.value],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List all private channels",
        user_message="List all private channels",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [ConversationType.PRIVATE_CHANNEL.value],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List all public and private channels",
        user_message="List all public and private channels",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [
                        ConversationType.PUBLIC_CHANNEL.value,
                        ConversationType.PRIVATE_CHANNEL.value,
                    ],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List direct message channels",
        user_message="List direct message channels",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [
                        ConversationType.DIRECT_MESSAGE.value,
                    ],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List group direct message channels",
        user_message="List group direct message channels",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [
                        ConversationType.MULTI_PERSON_DIRECT_MESSAGE.value,
                    ],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    suite.add_case(
        name="List my multi-person conversations",
        user_message="List my multi-person conversations",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_conversations,
                args={
                    "conversation_types": [
                        ConversationType.MULTI_PERSON_DIRECT_MESSAGE.value,
                    ],
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="conversation_types", weight=1.0),
        ],
    )

    return suite



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool

from arcade_google_hotels.enums import GoogleHotelsSortBy
from arcade_google_hotels.utils import call_serpapi, prepare_params


@tool(requires_secrets=["SERP_API_KEY"])
async def search_hotels(
    context: ToolContext,
    location: Annotated[str, "Location to search for hotels, e.g., a city name, a state, etc."],
    check_in_date: Annotated[str, "Check-in date in YYYY-MM-DD format"],
    check_out_date: Annotated[str, "Check-out date in YYYY-MM-DD format"],
    query: Annotated[
        str | None, "Anything that would be used in a regular Google Hotels search"
    ] = None,
    currency: Annotated[str | None, "Currency code for prices. Defaults to 'USD'"] = "USD",
    min_price: Annotated[int | None, "Minimum price per night. Defaults to no minimum"] = None,
    max_price: Annotated[int | None, "Maximum price per night. Defaults to no maximum"] = None,
    num_adults: Annotated[int | None, "Number of adults per room. Defaults to 2"] = 2,
    num_children: Annotated[int | None, "Number of children per room. Defaults to 0"] = 0,
    sort_by: Annotated[
        GoogleHotelsSortBy, "The sorting order of the results. Defaults to RELEVANCE"
    ] = GoogleHotelsSortBy.RELEVANCE,
    num_results: Annotated[
        int | None, "Maximum number of results to return. Defaults to 5. Max 20"
    ] = 5,
) -> Annotated[dict[str, Any], "Hotel search results from the Google Hotels API"]:
    """Retrieve hotel search results using the Google Hotels API."""
    # Prepare the request
    params = prepare_params(
        "google_hotels",
        q=f"{query}, {location}" if query else location,
        check_in_date=check_in_date,
        check_out_date=check_out_date,
        currency=currency,
        min_price=min_price,
        max_price=max_price,
        adults=num_adults,
        children=num_children,
        sort_by=sort_by.to_api_value(),
    )

    # Execute the request
    results = call_serpapi(context, params)

    # Parse the results
    properties = results.get("properties", [])[:num_results]

    # Remove unwanted fields from each property
    for hotel in properties:
        hotel.pop("images", None)
        hotel.pop("extracted_hotel_class", None)
        hotel.pop("reviews_breakdown", None)
        hotel.pop("serpapi_property_details_link", None)

    return {"properties": properties}



import json

from arcade_tdk.errors import RetryableToolError

from arcade_google_news.google_data import COUNTRY_CODES, LANGUAGE_CODES


class GoogleRetryableError(RetryableToolError):
    pass


class CountryNotFoundError(GoogleRetryableError):
    def __init__(self, country: str | None) -> None:
        valid_countries = json.dumps(COUNTRY_CODES, default=str)
        message = f"Country not found: '{country}'."
        additional_message = f"Valid countries are: {valid_countries}"
        super().__init__(message, additional_prompt_content=additional_message)


class LanguageNotFoundError(GoogleRetryableError):
    def __init__(self, language: str | None) -> None:
        valid_languages = json.dumps(LANGUAGE_CODES, default=str)
        message = f"Language not found: '{language}'."
        additional_message = f"Valid languages are: {valid_languages}"
        super().__init__(message, additional_prompt_content=additional_message)



from unittest.mock import AsyncMock, patch

import pytest
from arcade_tdk.errors import ToolExecutionError
from httpx import Response

from arcade_github.tools.models import RepoType
from arcade_github.tools.repositories import (
    count_stargazers,
    get_repository,
    list_org_repositories,
    list_repository_activities,
    list_review_comments_in_a_repository,
)


@pytest.fixture
def mock_context():
    context = AsyncMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.fixture
def mock_client():
    with patch("arcade_github.tools.repositories.httpx.AsyncClient") as client:
        yield client.return_value.__aenter__.return_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status_code,error_message,expected_error",
    [
        (422, "Validation Failed", "Error accessing.*: Validation failed"),
        (301, "Moved Permanently", "Error accessing.*: Moved permanently"),
        (404, "Not Found", "Error accessing.*: Resource not found"),
        (503, "Service Unavailable", "Error accessing.*: Service unavailable"),
        (410, "Gone", "Error accessing.*: Gone"),
    ],
)
async def test_error_responses(
    mock_context, mock_client, status_code, error_message, expected_error
):
    mock_client.get.return_value = Response(status_code, json={"message": error_message})
    mock_client.post.return_value = Response(status_code, json={"message": error_message})

    with pytest.raises(ToolExecutionError, match=expected_error):
        if status_code == 422:
            await list_org_repositories(mock_context, "org", repo_type=RepoType.ALL)
        elif status_code == 301:
            await count_stargazers(mock_context, "owner", "repo")
        elif status_code == 404:
            await list_org_repositories(mock_context, "non_existent_org")
        elif status_code == 503:
            await get_repository(mock_context, "owner", "repo")
        elif status_code == 410:
            await list_review_comments_in_a_repository(mock_context, "owner", "repo")


@pytest.mark.asyncio
async def test_list_repository_activities_invalid_cursor(mock_context, mock_client):
    mock_client.get.return_value = Response(422, json={"message": "Validation Failed"})

    with pytest.raises(ToolExecutionError, match="Error accessing.*: Validation failed"):
        await list_repository_activities(mock_context, "owner", "repo", before="invalid_cursor")


@pytest.mark.asyncio
async def test_count_stargazers_success(mock_context, mock_client):
    mock_client.get.return_value = Response(200, json={"stargazers_count": 42})

    result = await count_stargazers(mock_context, "owner", "repo")
    assert result == 42



from unittest.mock import MagicMock

import httpx
import pytest
from arcade_tdk.errors import ToolExecutionError

from arcade_x.tools.users import lookup_single_user_by_username


@pytest.mark.asyncio
async def test_lookup_single_user_by_username_success(tool_context, mock_httpx_client):
    """Test successful lookup of a user by username."""
    # Mock response for a successful user lookup
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "data": {
            "id": "1234567890",
            "name": "Test User",
            "username": "testuser",
            "description": "This is a test user",
            # Additional fields can be added here as needed
        }
    }
    mock_httpx_client.get.return_value = mock_response

    username = "testuser"
    result = await lookup_single_user_by_username(tool_context, username)

    assert "data" in result
    assert result["data"]["username"] == "testuser"
    assert result["data"]["name"] == "Test User"
    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_lookup_single_user_by_username_user_not_found(tool_context, mock_httpx_client):
    """Test behavior when looking up user fails due to API error"""
    # Mock response for user not found
    mock_response = httpx.HTTPStatusError(
        "Not Found", request=MagicMock(), response=MagicMock(status_code=404)
    )
    mock_httpx_client.get.side_effect = mock_response

    username = "nonexistentuser"
    with pytest.raises(ToolExecutionError):
        await lookup_single_user_by_username(tool_context, username)

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_lookup_single_user_by_username_api_error(tool_context, mock_httpx_client):
    """Test behavior when API returns an error other than 404."""
    # Mock response for API error
    mock_response = httpx.HTTPStatusError(
        "Internal Server Error", request=MagicMock(), response=MagicMock(status_code=500)
    )
    mock_httpx_client.get.side_effect = mock_response

    username = "testuser"
    with pytest.raises(ToolExecutionError):
        await lookup_single_user_by_username(tool_context, username)

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_lookup_single_user_by_username_network_error(tool_context, mock_httpx_client):
    """Test behavior when there is a network error during the request."""
    # Mock client.get to raise an HTTPError
    mock_httpx_client.get.side_effect = httpx.HTTPError("Network Error")

    username = "testuser"
    with pytest.raises(ToolExecutionError):
        await lookup_single_user_by_username(tool_context, username)

    mock_httpx_client.get.assert_called_once()



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_e2b
from arcade_e2b.enums import E2BSupportedLanguage
from arcade_e2b.tools.create_chart import create_static_matplotlib_chart
from arcade_e2b.tools.run_code import run_code

merge_sort_code = """
def merge_sort(arr):
    if len(arr) <= 1:
        return arr

    mid = len(arr) // 2
    left = merge_sort(arr[:mid])
    right = merge_sort(arr[mid:])

    return merge(left, right)

def merge(left, right):
    result = []
    i, j = 0, 0

    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            result.append(left[i])
            i += 1
        else:
            result.append(right[j])
            j += 1

    result.extend(left[i:])
    result.extend(right[j:])

    return result

sample_list = ["banana", "apple", "cherry", "date", "elderberry"]

sorted_list = merge_sort(sample_list)
print("Sorted list:", sorted_list)
"""

matplotlib_chart_code = """
import matplotlib.pyplot as plt

labels = ['Apples', 'Bananas', 'Cherries', 'Dates']
sizes = [30, 25, 20, 25]
colors = ['red', 'yellow', 'purple', 'brown']

plt.pie(sizes, labels=labels, colors=colors, autopct='%1.1f%%', startangle=90)

plt.axis('equal')

plt.title('Fruit Distribution')

plt.savefig('fruit_pie_chart.png')
"""

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_e2b)


@tool_eval()
def e2b_eval_suite():
    suite = EvalSuite(
        name="E2B Tools Evaluation",
        system_message="You are an AI assistant with access to E2B tools. Use them to help the user with their tasks.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Run code",
        user_message=f"Can you please run my merge sort algo?\n\n{merge_sort_code}",
        expected_tool_calls=[
            ExpectedToolCall(
                func=run_code,
                args={
                    "code": merge_sort_code,
                    "language": E2BSupportedLanguage.PYTHON,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="code", weight=0.8),
            BinaryCritic(critic_field="language", weight=0.2),
        ],
    )

    suite.add_case(
        name="Create static matplotlib chart",
        user_message=f"Run this code:\n\n{matplotlib_chart_code}",
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_static_matplotlib_chart,
                args={
                    "code": matplotlib_chart_code,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="code", weight=1.0),
        ],
    )

    return suite



import json
from typing import Annotated

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import GitHub
from arcade_tdk.errors import RetryableToolError

from arcade_github.tools.models import (
    DiffSide,
    PRSortProperty,
    PRState,
    ReviewCommentSortProperty,
    ReviewCommentSubjectType,
    SortDirection,
)
from arcade_github.tools.utils import (
    get_github_diff_headers,
    get_github_json_headers,
    get_url,
    handle_github_response,
    remove_none_values,
)


# Implements https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#list-pull-requests
# Example `arcade chat` usage:
#   "get all open PRs that <USER> has that are in the <OWNER>/<REPO> repo"
# TODO: Validate owner/repo combination is valid for the authenticated user.
#       If not, return RetryableToolError with available repos.
# TODO: list repo's branches and validate base is in the list (or default to main).
#       If not, return RetryableToolError with available branches.
@tool(requires_auth=GitHub())
async def list_pull_requests(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    state: Annotated[PRState | None, "The state of the pull requests to return."] = PRState.OPEN,
    head: Annotated[
        str | None,
        "Filter pulls by head user or head organization and branch name in the format of "
        "user:ref-name or organization:ref-name.",
    ] = None,
    base: Annotated[str | None, "Filter pulls by base branch name."] = "main",
    sort: Annotated[
        PRSortProperty | None, "The property to sort the results by."
    ] = PRSortProperty.CREATED,
    direction: Annotated[SortDirection | None, "The direction of the sort."] = None,
    per_page: Annotated[int, "The number of results per page (max 100)."] = 30,
    page: Annotated[int, "The page number of the results to fetch."] = 1,
    include_extra_data: Annotated[
        bool,
        "If true, return all the data available about the pull requests. "
        "This is a large payload and may impact performance - use with caution.",
    ] = False,
) -> Annotated[str, "JSON string containing a list of pull requests with their details"]:
    """
    List pull requests in a GitHub repository.

    Example:
    ```
    list_pull_requests(owner="octocat", repo="Hello-World", state=PRState.OPEN, sort=PRSort.UPDATED)
    ```
    """
    url = get_url("repo_pulls", owner=owner, repo=repo)
    params = {
        "base": base,
        "state": state.value if state else None,
        "sort": sort.value if sort else None,
        "per_page": min(max(1, per_page), 100),  # clamp per_page to 1-100
        "page": page,
        "head": head,
        "direction": direction,  # defaults to desc when sort is 'created'/'not specified', else asc
    }
    params = remove_none_values(params)
    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params)

    handle_github_response(response, url)

    pull_requests = response.json()
    results = []
    for pr in pull_requests:
        if include_extra_data:
            results.append(pr)
            continue
        results.append({
            "number": pr.get("number"),
            "title": pr.get("title"),
            "body": pr.get("body"),
            "state": pr.get("state"),
            "html_url": pr.get("html_url"),
            "diff_url": pr.get("diff_url"),
            "created_at": pr.get("created_at"),
            "updated_at": pr.get("updated_at"),
            "user": pr.get("user", {}).get("login"),
            "base": pr.get("base", {}).get("ref"),
            "head": pr.get("head", {}).get("ref"),
        })
    return json.dumps({"pull_requests": results})


# Implements https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#get-a-pull-request
# Example `arcade chat` usage:
#   "get the PR #72 in the <OWNER>/<REPO> repo. Include diff content in your response."
@tool(requires_auth=GitHub())
async def get_pull_request(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    include_diff_content: Annotated[
        bool | None,
        "If true, return the diff content of the pull request.",
    ] = False,
    include_extra_data: Annotated[
        bool | None,
        "If true, return all the data available about the pull requests. "
        "This is a large payload and may impact performance - use with caution.",
    ] = False,
) -> Annotated[
    str,
    "JSON string containing details of the specified pull request, "
    "optionally including diff content",
]:
    """
    Get details of a pull request in a GitHub repository.

    Example:
    ```
    get_pull_request(owner="octocat", repo="Hello-World", pull_number=1347)
    ```
    """
    url = get_url("repo_pull", owner=owner, repo=repo, pull_number=pull_number)
    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    diff_headers = get_github_diff_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers)
        if include_diff_content:
            diff_response = await client.get(url, headers=diff_headers)

    handle_github_response(response, url)

    if include_diff_content:
        handle_github_response(diff_response, url)

    pr_data = response.json()

    if include_extra_data:
        result = pr_data
        if include_diff_content:
            result["diff_content"] = diff_response.content.decode("utf-8")
        return json.dumps(result)

    important_info = {
        "number": pr_data.get("number"),
        "title": pr_data.get("title"),
        "body": pr_data.get("body"),
        "state": pr_data.get("state"),
        "html_url": pr_data.get("html_url"),
        "diff_url": pr_data.get("diff_url"),
        "created_at": pr_data.get("created_at"),
        "updated_at": pr_data.get("updated_at"),
        "user": pr_data.get("user", {}).get("login"),
        "base": pr_data.get("base", {}).get("ref"),
        "head": pr_data.get("head", {}).get("ref"),
    }

    if include_diff_content:
        important_info["diff_content"] = diff_response.content.decode("utf-8")

    return json.dumps(important_info)


# Implements https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#update-a-pull-request
# Example `arcade chat` usage:
#   "update PR #72 in the <OWNER>/<REPO> repo by changing the title to 'New Title' and
#   setting the body to 'This PR description was added via arcade chat!'."
# TODO: Enable this tool to append to the PR contents instead of only replacing content.
@tool(requires_auth=GitHub())
async def update_pull_request(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    title: Annotated[str | None, "The title of the pull request."] = None,
    body: Annotated[str | None, "The contents of the pull request."] = None,
    state: Annotated[PRState | None, "State of this Pull Request. Either open or closed."] = None,
    base: Annotated[str | None, "The name of the branch you want your changes pulled into."] = None,
    maintainer_can_modify: Annotated[
        bool | None, "Indicates whether maintainers can modify the pull request."
    ] = None,
) -> Annotated[str, "JSON string containing updated information about the pull request"]:
    """
    Update a pull request in a GitHub repository.

    Example:
    ```
    update_pull_request(
        owner="octocat",
        repo="Hello-World",
        pull_number=1347,
        title="new title",
        body="updated body",
    )
    ```
    """
    url = get_url("repo_pull", owner=owner, repo=repo, pull_number=pull_number)

    data = {
        "title": title,
        "body": body,
        "state": state.value if state else None,
        "base": base,
        "maintainer_can_modify": maintainer_can_modify,
    }
    data = remove_none_values(data)

    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.patch(url, headers=headers, json=data)

    handle_github_response(response, url)

    pr_data = response.json()
    important_info = {
        "url": pr_data.get("url"),
        "id": pr_data.get("id"),
        "html_url": pr_data.get("html_url"),
        "number": pr_data.get("number"),
        "state": pr_data.get("state"),
        "title": pr_data.get("title"),
        "user": pr_data.get("user", {}).get("login"),
        "body": pr_data.get("body"),
        "created_at": pr_data.get("created_at"),
        "updated_at": pr_data.get("updated_at"),
    }
    return json.dumps(important_info)


# Implements https://docs.github.com/en/rest/pulls/commits?apiVersion=2022-11-28#list-commits-on-a-pull-request
# Example `arcade chat` usage: "list all of the commits for the PR 72 in the <OWNER>/<REPO> repo"
@tool(requires_auth=GitHub())
async def list_pull_request_commits(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    per_page: Annotated[int, "The number of results per page (max 100)."] = 30,
    page: Annotated[int, "The page number of the results to fetch."] = 1,
    include_extra_data: Annotated[
        bool,
        "If true, return all the data available about the pull requests. "
        "This is a large payload and may impact performance - use with caution.",
    ] = False,
) -> Annotated[str, "JSON string containing a list of commits for the specified pull request"]:
    """
    List commits (from oldest to newest) on a pull request in a GitHub repository.

    Example:
    ```
    list_pull_request_commits(owner="octocat", repo="Hello-World", pull_number=1347)
    ```
    """
    url = get_url("repo_pull_commits", owner=owner, repo=repo, pull_number=pull_number)

    params = {
        "per_page": max(1, min(100, per_page)),  # clamp per_page to 1-100
        "page": page,
    }

    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params)

    handle_github_response(response, url)

    commits = response.json()
    if include_extra_data:
        return json.dumps({"commits": commits})

    filtered_commits = []
    for commit in commits:
        filtered_commit = {
            "sha": commit.get("sha"),
            "html_url": commit.get("html_url"),
            "diff_url": commit.get("html_url") + ".diff" if commit.get("html_url") else None,
            "commit": {
                "message": commit.get("commit", {}).get("message"),
                "author": commit.get("commit", {}).get("author", {}).get("name"),
                "committer": commit.get("commit", {}).get("committer", {}).get("name"),
                "date": commit.get("commit", {}).get("committer", {}).get("date"),
            },
            "author": commit.get("author", {}).get("login"),
            "committer": commit.get("committer", {}).get("login"),
        }
        filtered_commits.append(filtered_commit)

    return json.dumps({"commits": filtered_commits})


# Implements https://docs.github.com/en/rest/pulls/comments?apiVersion=2022-11-28#create-a-reply-for-a-review-comment
# Example `arcade chat` usage:
#   "create a reply to the review comment 1778019974 in arcadeai/arcade-ai for
#   the PR 72 that says 'Thanks for the suggestion.'"
# Note: This tool requires the ID of the review comment to reply to. To obtain this ID,
#       you should first call the `list_review_comments_on_pull_request` function.
#       The returned JSON will contain the `id` field for each comment, which can be used as the
#       `comment_id` parameter in this function.
@tool(requires_auth=GitHub())
async def create_reply_for_review_comment(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    comment_id: Annotated[int, "The unique identifier of the comment."],
    body: Annotated[str, "The text of the review comment."],
) -> Annotated[str, "JSON string containing details of the created reply comment"]:
    """
    Create a reply to a review comment for a pull request.

    Example:
    ```
    create_reply_for_review_comment(
        owner="octocat",
        repo="Hello-World",
        pull_number=1347,
        comment_id=42,
        body="Looks good to me!",
    )
    ```
    """
    url = get_url(
        "repo_pull_comment_replies",
        owner=owner,
        repo=repo,
        pull_number=pull_number,
        comment_id=comment_id,
    )

    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    data = {"body": body}

    async with httpx.AsyncClient() as client:
        response = await client.post(url, headers=headers, json=data)

    handle_github_response(response, url)

    return json.dumps(response.json())


# Implements https://docs.github.com/en/rest/pulls/comments?apiVersion=2022-11-28#list-review-comments-on-a-pull-request
# Example `arcade chat` usage: "list all of the review comments for PR 72 in <OWNER>/<REPO>"
@tool(requires_auth=GitHub())
async def list_review_comments_on_pull_request(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    sort: Annotated[
        ReviewCommentSortProperty | None,
        "The property to sort the results by. Can be one of: created, updated.",
    ] = ReviewCommentSortProperty.CREATED,
    direction: Annotated[
        SortDirection | None, "The direction to sort results. Can be one of: asc, desc."
    ] = SortDirection.DESC,
    since: Annotated[
        str | None,
        "Only show results that were last updated after the given time. "
        "This is a timestamp in ISO 8601 format: YYYY-MM-DDTHH:MM:SSZ.",
    ] = None,
    per_page: Annotated[int, "The number of results per page (max 100)."] = 30,
    page: Annotated[int, "The page number of the results to fetch."] = 1,
    include_extra_data: Annotated[
        bool,
        "If true, return all the data available about the review comments. "
        "This is a large payload and may impact performance - use with caution.",
    ] = False,
) -> Annotated[
    str, "JSON string containing a list of review comments for the specified pull request"
]:
    """
    List review comments on a pull request in a GitHub repository.

    Example:
    ```
    list_review_comments_on_pull_request(owner="octocat", repo="Hello-World", pull_number=1347)
    ```
    """
    url = get_url("repo_pull_comments", owner=owner, repo=repo, pull_number=pull_number)

    params = {
        "sort": sort,
        "direction": direction,
        "per_page": max(1, min(100, per_page)),  # clamp per_page to 1-100
        "page": page,
        "since": since,
    }
    params = remove_none_values(params)

    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params)

    handle_github_response(response, url)

    review_comments = response.json()
    if include_extra_data:
        return json.dumps(review_comments)

    filtered_comments = []
    for comment in review_comments:
        filtered_comment = {
            "id": comment.get("id"),
            "url": comment.get("url"),
            "diff_hunk": comment.get("diff_hunk"),
            "path": comment.get("path"),
            "position": comment.get("position"),
            "original_position": comment.get("original_position"),
            "commit_id": comment.get("commit_id"),
            "original_commit_id": comment.get("original_commit_id"),
            "in_reply_to_id": comment.get("in_reply_to_id"),
            "user": comment.get("user", {}).get("login"),
            "body": comment.get("body"),
            "created_at": comment.get("created_at"),
            "updated_at": comment.get("updated_at"),
            "html_url": comment.get("html_url"),
            "line": comment.get("line"),
            "side": comment.get("side"),
            "pull_request_url": comment.get("pull_request_url"),
        }
        filtered_comments.append(filtered_comment)
    return json.dumps({"review_comments": filtered_comments})


# Implements https://docs.github.com/en/rest/pulls/comments?apiVersion=2022-11-28#create-a-review-comment-for-a-pull-request
# Example `arcade chat` usage: "create a review comment for PR 72 in <OWNER>/<REPO> that says
#           'Great stuff! This looks good to merge. Add the comment to README.md file.'"
# TODO: Verify that path parameter exists in the PR's files that have changed
#       (Or should we allow for any file in the repo?).
#       If not, then throw RetryableToolError with all valid file paths.
@tool(requires_auth=GitHub())
async def create_review_comment(
    context: ToolContext,
    owner: Annotated[str, "The account owner of the repository. The name is not case sensitive."],
    repo: Annotated[
        str,
        "The name of the repository without the .git extension. The name is not case sensitive.",
    ],
    pull_number: Annotated[int, "The number that identifies the pull request."],
    body: Annotated[str, "The text of the review comment."],
    path: Annotated[str, "The relative path to the file that necessitates a comment."],
    commit_id: Annotated[
        str | None,
        "The SHA of the commit needing a comment. If not provided, the latest commit SHA of the "
        "PR's base branch will be used.",
    ] = None,
    start_line: Annotated[
        int | None,
        "The start line of the range of lines in the pull request diff that the "
        "comment applies to. Required unless 'subject_type' is 'file'.",
    ] = None,
    end_line: Annotated[
        int | None,
        "The end line of the range of lines in the pull request diff that the "
        "comment applies to. Required unless 'subject_type' is 'file'.",
    ] = None,
    side: Annotated[
        DiffSide | None,
        "The side of the diff that the pull request's changes appear on. "
        "Use LEFT for deletions that appear in red. Use RIGHT for additions that appear in green "
        "or unchanged lines that appear in white and are shown for context",
    ] = DiffSide.RIGHT,
    start_side: Annotated[
        str | None, "The starting side of the diff that the comment applies to."
    ] = None,
    subject_type: Annotated[
        ReviewCommentSubjectType | None,
        "The type of subject that the comment applies to. Can be one of: file, hunk, or line.",
    ] = ReviewCommentSubjectType.FILE,
    include_extra_data: Annotated[
        bool,
        "If true, return all the data available about the review comment. "
        "This is a large payload and may impact performance - use with caution.",
    ] = False,
) -> Annotated[str, "JSON string containing details of the created review comment"]:
    """
    Create a review comment for a pull request in a GitHub repository.

    If the subject_type is not 'file', then the start_line and end_line parameters are required.
    If the subject_type is 'file', then the start_line and end_line parameters are ignored.
    If the commit_id is not provided, the latest commit SHA of the PR's base branch will be used.

    Example:
    ```
    create_review_comment(
        owner="octocat",
        repo="Hello-World",
        pull_number=1347,
        body="Great stuff!",
        commit_id="6dcb09b5b57875f334f61aebed695e2e4193db5e",
        path="file1.txt",
        line=2,
        side="RIGHT"
    )
    ```
    """
    # If the subject_type is 'file', then the line_range parameter is ignored
    if subject_type == ReviewCommentSubjectType.FILE:
        start_line, end_line = None, None

    if (start_line is None or end_line is None) and subject_type != ReviewCommentSubjectType.FILE:
        message = (
            "'start_line' and 'end_line' parameters are required when 'subject_type' "
            "parameter is not 'file'. Either provide both a start_line and end_line or set "
            "subject_type to 'file'."
        )
        raise RetryableToolError(
            message=message,
            developer_message=message,
            additional_prompt_content=message,
        )

    # Ensure the line range goes from lowest to highest
    if start_line is not None and end_line is not None:
        start_line, end_line = (min(start_line, end_line), max(start_line, end_line))

    # Get the latest commit SHA of the PR's base branch and use that for the commit_id
    if not commit_id:
        commits_json = await list_pull_request_commits(context, owner, repo, pull_number)
        commits_data = json.loads(commits_json)
        commits = commits_data.get("commits", [])
        latest_commit = commits[-1] if commits else {}
        commit_id = latest_commit.get("sha")

    if not commit_id:
        message = (
            f"Failed to get the latest commit SHA of PR {pull_number} in repo {repo} owned by "
            "{owner}. Does the PR exist?"
        )
        raise RetryableToolError(
            message=message,
            developer_message=message,
            additional_prompt_content=message,
        )

    url = get_url("repo_pull_comments", owner=owner, repo=repo, pull_number=pull_number)
    data = {
        "body": body,
        "commit_id": commit_id,
        "path": path,
        "side": side,
        "line": end_line if end_line else None,
        "start_line": start_line
        if start_line and start_line != end_line
        else None,  # Only send start_line when using multi-line comments
        "start_side": start_side,
    }
    data = remove_none_values(data)
    headers = get_github_json_headers(
        context.authorization.token if context.authorization and context.authorization.token else ""
    )

    async with httpx.AsyncClient() as client:
        response = await client.post(url, headers=headers, json=data)

    handle_github_response(response, url)

    comment_data = response.json()
    if include_extra_data:
        return json.dumps(comment_data)

    important_info = {
        "id": comment_data.get("id"),
        "url": comment_data.get("url"),
        "body": comment_data.get("body"),
        "path": comment_data.get("path"),
        "line": comment_data.get("line"),
        "side": comment_data.get("side"),
        "commit_id": comment_data.get("commit_id"),
        "user": comment_data.get("user", {}).get("login"),
        "created_at": comment_data.get("created_at"),
        "updated_at": comment_data.get("updated_at"),
        "html_url": comment_data.get("html_url"),
    }
    return json.dumps(important_info)



from unittest.mock import MagicMock, patch

import pytest
from arcade_cli.constants import PROD_ENGINE_HOST
from arcade_cli.main import cli
from typer.testing import CliRunner

runner = CliRunner()


@pytest.mark.parametrize(
    "args, expected_url",
    [
        ([], f"https://{PROD_ENGINE_HOST}/dashboard"),
        (["--local"], "http://localhost:9099/dashboard"),
        (["--host", "custom.host.com"], "https://custom.host.com/dashboard"),
        (["-h", "api.arcade.dev", "-p", "9099"], "https://api.arcade.dev:9099/dashboard"),
        (["--local", "--port", "9099"], "http://localhost:9099/dashboard"),
        (["--local", "--tls"], "https://localhost:9099/dashboard"),
        (["--no-tls"], f"http://{PROD_ENGINE_HOST}/dashboard"),
    ],
)
def test_dashboard_url_construction(args, expected_url):
    """Test that the dashboard command constructs the correct URL with various args."""
    with (
        patch("webbrowser.open") as mock_open,
        patch("arcade_cli.main.validate_and_get_config") as mock_validate,
        patch("arcade_cli.main.log_engine_health") as mock_health_check,
    ):
        # Setup mocks
        mock_open.return_value = True  # Successfully opened browser
        mock_validate.return_value = MagicMock()
        mock_health_check.return_value = None  # Successful health check

        # Run command
        result = runner.invoke(cli, ["dashboard", *args])

        assert result.exit_code == 0
        mock_open.assert_called_once_with(expected_url)
        mock_health_check.assert_called_once()


def test_fallback_when_browser_fails():
    """Test fallback message when browser.open fails."""
    with (
        patch("webbrowser.open") as mock_open,
        patch("arcade_cli.main.validate_and_get_config") as mock_validate,
        patch("arcade_cli.main.log_engine_health") as mock_health_check,
        patch("arcade_cli.main.console.print") as mock_print,
    ):
        mock_open.return_value = False  # Failed to open browser
        mock_validate.return_value = MagicMock()
        mock_health_check.return_value = None

        result = runner.invoke(cli, ["dashboard"])

        assert result.exit_code == 0
        mock_print.assert_any_call(
            f"If a browser doesn't open automatically, copy this URL and paste it into your browser: https://{PROD_ENGINE_HOST}/dashboard",
            style="dim",
        )


def test_health_check_success():
    """Test successful health check."""
    with (
        patch("webbrowser.open") as mock_open,
        patch("arcade_cli.main.validate_and_get_config") as mock_validate,
        patch("arcade_cli.main.log_engine_health") as mock_health_check,
    ):
        mock_open.return_value = True
        mock_validate.return_value = MagicMock()
        mock_health_check.return_value = None  # Successful health check

        result = runner.invoke(cli, ["dashboard"])

        assert result.exit_code == 0
        mock_health_check.assert_called_once()
        mock_open.assert_called_once()



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_search
from arcade_search.constants import (
    DEFAULT_GOOGLE_MAPS_COUNTRY,
    DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
    DEFAULT_GOOGLE_MAPS_LANGUAGE,
    DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
)
from arcade_search.enums import GoogleMapsDistanceUnit, GoogleMapsTravelMode
from arcade_search.tools.google_maps import (
    get_directions_between_addresses,
    get_directions_between_coordinates,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)

catalog = ToolCatalog()
# Register the Google Search tool
catalog.add_module(arcade_search)


@tool_eval()
def google_maps_directions_by_addresses_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the Google Maps Directions tools."""
    suite = EvalSuite(
        name="Google Maps Directions Tool Evaluation",
        system_message="You are an AI assistant that can get directions from Google Maps using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get directions between two addresses",
        user_message="Get directions from Google Maps between the following addresses: 1600 Amphitheatre Parkway, Mountain View, CA 94043 and 1 Infinite Loop, Cupertino, CA 95014.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_addresses,
                args={
                    "origin_address": "1600 Amphitheatre Parkway, Mountain View, CA 94043",
                    "destination_address": "1 Infinite Loop, Cupertino, CA 95014",
                    "language": DEFAULT_GOOGLE_MAPS_LANGUAGE,
                    "country": DEFAULT_GOOGLE_MAPS_COUNTRY,
                    "distance_unit": DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
                    "travel_mode": DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_address", weight=0.3),
            SimilarityCritic(critic_field="destination_address", weight=0.3),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    suite.add_case(
        name="Get directions between two addresses with custom distance unit and travel mode",
        user_message="Get walking directions from Google Maps between the following addresses in miles: 1600 Amphitheatre Parkway, Mountain View, CA 94043 and 1 Infinite Loop, Cupertino, CA 95014.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_addresses,
                args={
                    "origin_address": "1600 Amphitheatre Parkway, Mountain View, CA 94043",
                    "destination_address": "1 Infinite Loop, Cupertino, CA 95014",
                    "language": DEFAULT_GOOGLE_MAPS_LANGUAGE,
                    "country": DEFAULT_GOOGLE_MAPS_COUNTRY,
                    "distance_unit": GoogleMapsDistanceUnit.MILES.value,
                    "travel_mode": GoogleMapsTravelMode.WALKING.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_address", weight=0.3),
            SimilarityCritic(critic_field="destination_address", weight=0.3),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    suite.add_case(
        name="Get directions between two addresses in a given country and language",
        user_message="Get directions from Google Maps in Portuguese between the following addresses in Brazil: Rua do Amendoim, 1, Belo Horizonte, MG and Av. do Descobrimento, 515, Porto Seguro, BA.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_addresses,
                args={
                    "origin_address": "Rua do Amendoim, 1, Belo Horizonte, MG",
                    "destination_address": "Av. do Descobrimento, 515, Porto Seguro, BA",
                    "language": "pt",
                    "country": "br",
                    "distance_unit": DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
                    "travel_mode": DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_address", weight=0.3),
            SimilarityCritic(critic_field="destination_address", weight=0.3),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    return suite


@tool_eval()
def google_maps_directions_by_coordinates_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the Google Maps Directions tools."""
    suite = EvalSuite(
        name="Google Maps Directions Tool Evaluation",
        system_message="You are an AI assistant that can get directions from Google Maps using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get directions between two coordinates",
        user_message="Get directions from Google Maps between the following coordinates: 37.422740,-122.084961 and 37.331820,-122.031180.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_coordinates,
                args={
                    "origin_latitude": "37.422740",
                    "origin_longitude": "-122.084961",
                    "destination_latitude": "37.331820",
                    "destination_longitude": "-122.031180",
                    "language": DEFAULT_GOOGLE_MAPS_LANGUAGE,
                    "country": DEFAULT_GOOGLE_MAPS_COUNTRY,
                    "distance_unit": DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
                    "travel_mode": DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_latitude", weight=0.15),
            SimilarityCritic(critic_field="origin_longitude", weight=0.15),
            SimilarityCritic(critic_field="destination_latitude", weight=0.15),
            SimilarityCritic(critic_field="destination_longitude", weight=0.15),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    suite.add_case(
        name="Get directions between two coordinates with custom distance unit and travel mode",
        user_message="Get walking directions from Google Maps between the following coordinates in miles: 37.422740,-122.084961 and 37.331820,-122.031180.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_coordinates,
                args={
                    "origin_latitude": "37.422740",
                    "origin_longitude": "-122.084961",
                    "destination_latitude": "37.331820",
                    "destination_longitude": "-122.031180",
                    "language": DEFAULT_GOOGLE_MAPS_LANGUAGE,
                    "country": DEFAULT_GOOGLE_MAPS_COUNTRY,
                    "distance_unit": GoogleMapsDistanceUnit.MILES.value,
                    "travel_mode": GoogleMapsTravelMode.WALKING.value,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_latitude", weight=0.15),
            SimilarityCritic(critic_field="origin_longitude", weight=0.15),
            SimilarityCritic(critic_field="destination_latitude", weight=0.15),
            SimilarityCritic(critic_field="destination_longitude", weight=0.15),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    suite.add_case(
        name="Get directions between two coordinates in a given country and language",
        user_message="Get directions from Google Maps in Portuguese between the following coordinates in Brazil: 37.422740,-122.084961 and 37.331820,-122.031180.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_directions_between_coordinates,
                args={
                    "origin_latitude": "37.422740",
                    "origin_longitude": "-122.084961",
                    "destination_latitude": "37.331820",
                    "destination_longitude": "-122.031180",
                    "language": "pt",
                    "country": "br",
                    "distance_unit": DEFAULT_GOOGLE_MAPS_DISTANCE_UNIT,
                    "travel_mode": DEFAULT_GOOGLE_MAPS_TRAVEL_MODE,
                },
            )
        ],
        critics=[
            SimilarityCritic(critic_field="origin_latitude", weight=0.15),
            SimilarityCritic(critic_field="origin_longitude", weight=0.15),
            SimilarityCritic(critic_field="destination_latitude", weight=0.15),
            SimilarityCritic(critic_field="destination_longitude", weight=0.15),
            BinaryCritic(critic_field="language", weight=0.1),
            BinaryCritic(critic_field="country", weight=0.1),
            BinaryCritic(critic_field="distance_unit", weight=0.1),
            BinaryCritic(critic_field="travel_mode", weight=0.1),
        ],
    )

    return suite



from unittest.mock import MagicMock

import httpx
import pytest
from arcade_tdk.errors import ToolExecutionError

from arcade_spotify.tools.tracks import get_track_from_id
from arcade_spotify.tools.utils import get_url


@pytest.mark.asyncio
async def test_get_track_from_id_success(tool_context, mock_httpx_client, sample_track):
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = sample_track
    mock_httpx_client.request.return_value = mock_response

    result = await get_track_from_id(tool_context, "1234567890")

    assert result == sample_track

    mock_httpx_client.request.assert_called_once_with(
        "GET",
        get_url("tracks_get_track", track_id="1234567890"),
        headers={"Authorization": f"Bearer {tool_context.authorization.token}"},
        params=None,
        json=None,
    )


@pytest.mark.asyncio
async def test_get_track_from_id_rate_limit_error(tool_context, mock_httpx_client):
    mock_response = MagicMock()
    mock_response = httpx.HTTPStatusError(
        "Too Many Requests", request=MagicMock(), response=MagicMock(status_code=429)
    )
    mock_httpx_client.request.side_effect = mock_response

    with pytest.raises(ToolExecutionError):
        await get_track_from_id(tool_context, "1234567890")



import os

DEFAULT_GOOGLE_LANGUAGE = os.getenv("ARCADE_GOOGLE_LANGUAGE", "en")
DEFAULT_GOOGLE_COUNTRY = os.getenv("ARCADE_GOOGLE_COUNTRY")
DEFAULT_GOOGLE_SHOPPING_LANGUAGE = os.getenv(
    "ARCADE_GOOGLE_SHOPPING_LANGUAGE", DEFAULT_GOOGLE_LANGUAGE
)
DEFAULT_GOOGLE_SHOPPING_COUNTRY = os.getenv(
    "ARCADE_GOOGLE_SHOPPING_COUNTRY", DEFAULT_GOOGLE_COUNTRY
)
